#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# sugarcane.py
# projects
#
# Created by Haibao Tang on 12/02/19
# Copyright © 2019 Haibao Tang. All rights reserved.
#
"""
Simulate sugarcane genomes and analyze the diversity in the progeny genomes.
"""

from collections import Counter, defaultdict
from dataclasses import dataclass
from enum import Enum
from itertools import combinations, groupby, product
from multiprocessing import Pool, cpu_count
import os.path as op
from random import randint, random
import sys
from typing import Dict, List, Tuple
from uuid import uuid4

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import seaborn as sns

from ..apps.base import ActionDispatcher, OptionParser, flatten, logger, mkdir
from ..formats.blast import Blast
from ..graphics.base import (
    Rectangle,
    adjust_spines,
    get_shades,
    markup,
    normalize_axes,
    savefig,
)
from ..graphics.chromosome import Chromosome as ChromosomePlot
from ..utils.cbook import short_float

SoColor = "#7436a4"  # Purple
SsColor = "#5a8340"  # Green
HAPLOID_GENE_COUNT = 160
SO_PLOIDY = 8
SS_PLOIDY = 16
SO_GENE_COUNT = SO_PLOIDY * HAPLOID_GENE_COUNT
SS_GENE_COUNT = SS_PLOIDY * HAPLOID_GENE_COUNT
SO_CHROM_COUNT = 10
SS_CHROM_COUNT = 8
CROSSES = ("F1", "BC1", "BC2", "BC3", "BC4")
TITLES = (
    (r"$\mathrm{F_1}$", None),
    (r"$\mathrm{BC_1}$", None),
    (r"$\mathrm{BC_2}$", None),
    (r"$\mathrm{BC_3}$", None),
    (r"$\mathrm{BC_4}$", None),
)


class CrossMode(Enum):
    """
    How the F1 is generated.
    """

    nplusn = "n+n"
    nx2plusn = "nx2+n"
    twoplusnFDR = "2n+n_FDR"
    twoplusnSDR = "2n+n_SDR"


@dataclass(frozen=True)
class Gene:
    """
    A gene in a chromosome.
    """

    chrom: str
    haplotype: str
    idx: int


@dataclass
class Chromosome(list):
    """
    A chromosome with genes.
    """

    subgenome: str
    chrom: str
    genes: List[Gene]
    uuid: str

    def __init__(self, subgenome: str, chrom: str, haplotype: str, gene_count: int):
        self.subgenome = subgenome
        self.chrom = chrom
        self.genes = [Gene(chrom, haplotype, i + 1) for i in range(gene_count)]
        self.uuid = str(uuid4())

    def __str__(self) -> str:
        """Merge genes into consecutive ranges."""
        ranges = []
        for haplotype, genes in groupby(self.genes, key=lambda x: x.haplotype):
            genes = list(genes)
            start, end = genes[0].idx, genes[-1].idx
            ranges.append(f"{haplotype}{start}-{end}")
        return f"{self.subgenome}-{self.chrom}|{','.join(ranges)}"

    def __len__(self) -> int:
        return len(self.genes)

    def duplicate(self) -> "Chromosome":
        """Duplicate the chromosome."""
        return Chromosome.make(self.subgenome, self.chrom, self.genes)

    @classmethod
    def make(cls, subgenome: str, chrom: str, genes: List[Gene]) -> "Chromosome":
        chrom = Chromosome(subgenome, chrom, "", 0)
        chrom.genes = genes
        return chrom

    def num_matching_genes(self, other: "Chromosome") -> int:
        """Count the number of matching genes between two chromosomes"""
        return sum(1 for x, y in zip(self.genes, other.genes) if x == y)

    def truncate_n(self, n: int):
        """Truncate the chromosome to n genes."""
        self.genes = self.genes[:n]

    def truncate(self, ratio: float):
        """Truncate the chromosome to a ratio of genes."""
        before = str(self)
        n = round(len(self) * ratio)
        self.truncate_n(n)
        logger.info("Truncate %s to %s (%d)", before, str(self), n)


# Simulate genome composition
class Genome:
    """
    A genome with a collection of chromosomes.
    """

    name: str
    chromosomes: List[Chromosome]

    def __init__(
        self,
        name: str,
        subgenome: str,
        ploidy: int,
        haploid_chromosome_count: int,
        haploid_gene_count: int,
    ):
        """
        Simulate a genome with given ploidy and haploid_chromosome_count. Example:

        >>> print(Genome("test", "pf", 2, 3, 90))
        test: pf-chr01|a1-30;pf-chr01|b1-30;pf-chr02|a1-30;pf-chr02|b1-30;pf-chr03|a1-30;pf-chr03|b1-30
        """
        self.name = name
        chrom_gene_count = (
            haploid_gene_count // haploid_chromosome_count
            if haploid_chromosome_count
            else 0
        )
        chromosomes = []
        for i in range(haploid_chromosome_count):
            chromosomes += [
                Chromosome(
                    subgenome, f"chr{i + 1:02d}", chr(ord("a") + j), chrom_gene_count
                )
                for j in range(ploidy)
            ]
        self.chromosomes = chromosomes

    def __str__(self):
        return self.name + ": " + ";".join(str(_) for _ in self.chromosomes)

    def ploidy(self, chrom: str) -> int:
        """
        Return the ploidy of a chromosome.
        """
        return sum(1 for x in self.chromosomes if x.chrom == chrom)

    @classmethod
    def from_str(cls, s: str) -> "Genome":
        """
        Parse a string representation of a genome.

        >>> s = "test: pf-chr01|a1-30;pf-chr01|b1-30;pf-chr02|a1-30;pf-chr02|b1-30;pf-chr03|a1-30;pf-chr03|b1-30"
        >>> str(Genome.from_str(s)) == s
        True
        """
        name, chroms = s.split(": ")
        genome = Genome(name, "", 0, 0, 0)
        chromosomes = []
        for x in chroms.split(";"):
            chrom, genes = x.split("|")
            subgenome, chrom = chrom.split("-")
            cgenes = []
            for genes in genes.split(","):
                haplotype = genes[0]
                start, end = [int(x) for x in genes[1:].split("-")]
                cgenes += [Gene(chrom, haplotype, i) for i in range(start, end + 1)]
            chrom = Chromosome.make(subgenome, chrom, cgenes)
            chromosomes.append(chrom)
        genome.chromosomes = chromosomes
        return genome

    @classmethod
    def make(cls, name: str, chromosomes: List[Chromosome]) -> "Genome":
        genome = Genome(name, "", 0, 0, 0)
        genome.chromosomes = sorted(chromosomes, key=genome._sort_key)
        return genome

    def _sort_key(self, x: Chromosome):
        return x.subgenome, x.chrom

    def pair_with_weights(self) -> List[List[str]]:
        """
        Pair chromosomes by similarity, results are sorted in descending similarity.
        """
        G = nx.Graph()
        scores = {}
        self.chromosomes.sort(key=self._sort_key)
        for _, chromosomes in groupby(self.chromosomes, key=self._sort_key):
            for a, b in combinations(chromosomes, 2):
                weight = a.num_matching_genes(b) + 1  # +1 so we don't have zero weight
                G.add_edge(a.uuid, b.uuid, weight=weight)
                scores[tuple(sorted((a.uuid, b.uuid)))] = weight
        # Find the maximum matching
        matching = nx.max_weight_matching(G)
        return sorted(matching, key=lambda x: scores[tuple(sorted(x))], reverse=True)

    def pair_chromosomes(
        self, inplace=False
    ) -> Tuple[List[List[Chromosome]], List[Chromosome]]:
        """
        Pair chromosomes by similarity.
        """
        matching = self.pair_with_weights()
        # Partition the chromosomes into paired and singleton
        paired = set()
        pairs = []
        for a, b in matching:
            paired.add(a)
            paired.add(b)
            pair = [x for x in self.chromosomes if x.uuid in (a, b)]
            pairs.append(pair)
        singletons = [x for x in self.chromosomes if x.uuid not in paired]
        if inplace:
            reordered = []
            for a, b in pairs:
                reordered += [a, b]
            reordered += singletons
            reordered.sort(key=self._sort_key)
            self.chromosomes = reordered
        return pairs, singletons

    def _crossover_chromosomes(
        self, a: Chromosome, b: Chromosome, sdr: bool
    ) -> List[Chromosome]:
        """
        Crossover two chromosomes.
        """
        if a.subgenome != b.subgenome or a.chrom != b.chrom or len(a) != len(b):
            raise ValueError("Incompatible chromosomes. Crossover failed.")
        subgenome, chrom = a.subgenome, a.chrom
        n = len(a)
        if random() < 0.5:
            a, b = b, a
        if random() < 0.5:
            crossover_point = randint(0, n // 2 - 1)
            recombinant = a.genes[:crossover_point] + b.genes[crossover_point:]
            non_recombinant = b.genes
        else:
            crossover_point = randint(n // 2, n - 1)
            recombinant = a.genes[:crossover_point] + b.genes[crossover_point:]
            non_recombinant = a.genes
        recombinant = Chromosome.make(subgenome, chrom, recombinant)
        non_recombinant = Chromosome.make(subgenome, chrom, non_recombinant)
        return [recombinant, non_recombinant] if sdr else [recombinant]

    def _gamete(self, sdr: bool):
        """Randomly generate a gamete from current genome."""
        gamete_chromosomes = []
        paired_chromosomes, singleton_chromosomes = self.pair_chromosomes()
        for a, b in paired_chromosomes:
            gamete_chromosomes += self._crossover_chromosomes(a, b, sdr=sdr)

        # Randomly assign the rest, singleton chromosomes
        for a in singleton_chromosomes:
            if random() < 0.5:
                gamete_chromosomes.append(a)

        tag = "sdr gamete" if sdr else "gamete"
        return Genome.make(f"{self.name} {tag}", gamete_chromosomes)

    @property
    def gamete(self):
        return self._gamete(sdr=False)

    @property
    def gamete_fdr(self):
        return Genome.make(f"{self.name} fdr gamete", self.chromosomes)

    @property
    def gamete_sdr(self):
        return self._gamete(sdr=True)

    def mate_nplusn(self, name: str, other_genome: "Genome", verbose: bool = True):
        if verbose:
            print(
                f"Crossing '{self.name}' x '{other_genome.name}' (n+n)", file=sys.stderr
            )
        f1_chromosomes = self.gamete.chromosomes + other_genome.gamete.chromosomes
        return Genome.make(name, f1_chromosomes)

    def mate_nx2plusn(self, name: str, other_genome: "Genome", verbose: bool = True):
        if verbose:
            print(
                f"Crossing '{self.name}' x '{other_genome.name}' (nx2+n)",
                file=sys.stderr,
            )
        gamete = self.gamete.chromosomes
        duplicate = [x.duplicate() for x in gamete]
        f1_chromosomes = gamete + duplicate + other_genome.gamete.chromosomes
        return Genome.make(name, f1_chromosomes)

    def mate_2nplusn_FDR(self, name: str, other_genome: "Genome", verbose: bool = True):
        if verbose:
            print(
                f"Crossing '{self.name}' x '{other_genome.name}' (2n+n_FDR)",
                file=sys.stderr,
            )
        f1_chromosomes = self.gamete_fdr.chromosomes + other_genome.gamete.chromosomes
        return Genome.make(name, f1_chromosomes)

    def mate_2nplusn_SDR(self, name: str, other_genome: "Genome", verbose: bool = True):
        if verbose:
            print(
                f"Crossing '{self.name}' x '{other_genome.name}' (2n+n_SDR)",
                file=sys.stderr,
            )
        f1_chromosomes = sorted(
            self.gamete_sdr.chromosomes + other_genome.gamete.chromosomes
        )
        return Genome.make(name, f1_chromosomes)

    @property
    def summary(self) -> List[Tuple[str, int, int, int]]:
        ans = []
        total_chrom_count = 0
        total_so_size = 0
        total_ss_size = 0

        self.chromosomes.sort(key=lambda x: x.subgenome)
        for subgenome, chromosomes in groupby(
            self.chromosomes, key=lambda x: x.subgenome
        ):
            chromosomes = list(chromosomes)
            uniq_genes = set(flatten(x.genes for x in chromosomes))
            group_count = len(uniq_genes)
            group_chrom_count = group_count / len(chromosomes[0])
            group_so_size = group_count if subgenome == "SO" else 0
            group_ss_size = group_count if subgenome == "SS" else 0
            group_size = group_so_size + group_ss_size
            ans.append(
                (
                    subgenome,
                    group_chrom_count,
                    group_so_size / group_size,
                    group_ss_size / group_size,
                )
            )
            total_chrom_count += group_chrom_count
            total_so_size += group_so_size
            total_ss_size += group_ss_size
        total_size = total_so_size + total_ss_size
        ans.append(
            (
                "Total",
                total_chrom_count,
                total_so_size / total_size,
                total_ss_size / total_size,
            )
        )
        return ans

    def print_summary(self):
        print("[SUMMARY]")
        for group, group_chrom_count, group_so_size, group_ss_size in self.summary:
            print(
                f"{group}: chrom_count={group_chrom_count}, SO={group_so_size}, SS={group_ss_size}"
            )

    def truncate_chromosome(self, subgenome: str, chrom: str, ratio: float):
        """
        Truncate the specified chromosome.
        """
        for c in self.chromosomes:
            if c.subgenome == subgenome and c.chrom == chrom:
                c.truncate(ratio)


class GenomeSummary:
    def __init__(self, SO_data, SS_data, percent_SO_data, percent_SS_data):
        self.SO_data = SO_data
        self.SS_data = SS_data
        self.percent_SO_data = percent_SO_data
        self.percent_SS_data = percent_SS_data

    def _summary(self, a, tag, precision=0) -> Tuple[str, str]:
        mean, mn, mx = (
            round(np.mean(a), precision),
            round(np.min(a), precision),
            round(np.max(a), precision),
        )
        s = f"*{tag}* chr: {mean:.0f}"
        if mn == mean and mx == mean:
            return s, ""
        return s, f" ({mn:.0f}-{mx:.0f})"

    def _percent_summary(self, a, tag, precision=1) -> Tuple[str, str]:
        mean, mn, mx = (
            round(np.mean(a), precision),
            round(np.min(a), precision),
            round(np.max(a), precision),
        )
        s = f"*{tag}*%: {short_float(mean, precision)}%"
        print(s)
        if mn == mean and mx == mean:
            return s, ""
        return s, f"({short_float(mn, precision)}-{short_float(mx, precision)}%)"

    @property
    def percent_SO_summary(self) -> Tuple[str, str]:
        return self._percent_summary(self.percent_SO_data, "So")

    @property
    def percent_SS_summary(self) -> Tuple[str, str]:
        return self._percent_summary(self.percent_SS_data, "Ss")

    @property
    def SO_summary(self) -> Tuple[str, str]:
        return self._summary(self.SO_data, "So")

    @property
    def SS_summary(self) -> Tuple[str, str]:
        return self._summary(self.SS_data, "Ss")


def simulate_F1(SO: Genome, SS: Genome, mode: CrossMode, verbose: bool = False):
    SO_SS_F1 = None
    if mode == CrossMode.nx2plusn:
        SO_SS_F1 = SO.mate_nx2plusn("SOxSS F1", SS, verbose=verbose)
    elif mode == CrossMode.twoplusnFDR:
        SO_SS_F1 = SO.mate_2nplusn_FDR("SOxSS F1", SS, verbose=verbose)
    elif mode == CrossMode.twoplusnSDR:
        SO_SS_F1 = SO.mate_2nplusn_SDR("SOxSS F1", SS, verbose=verbose)
    if verbose and SO_SS_F1:
        SO_SS_F1.print_summary()
    return SO_SS_F1


def simulate_BC1(SO: Genome, SS_SO_F1: Genome, mode: CrossMode, verbose=False):
    SS_SO_BC1 = None
    if mode == CrossMode.nx2plusn:
        SS_SO_BC1 = SO.mate_nx2plusn("SOxSS BC1", SS_SO_F1, verbose=verbose)
    elif mode == CrossMode.twoplusnFDR:
        SS_SO_BC1 = SO.mate_2nplusn_FDR("SOxSS BC1", SS_SO_F1, verbose=verbose)
    elif mode == CrossMode.twoplusnSDR:
        SS_SO_BC1 = SO.mate_2nplusn_SDR("SOxSS BC1", SS_SO_F1, verbose=verbose)
    return SS_SO_BC1


def simulate_F1_to_BC4(
    SO: Genome, SS: Genome, mode: CrossMode, verbose=False
) -> List[Genome]:
    SS_SO_F1 = simulate_F1(SO, SS, mode=mode, verbose=verbose)
    SS_SO_BC1 = simulate_BC1(SO, SS_SO_F1, mode=mode, verbose=verbose)
    SS_SO_BC2_nplusn = SO.mate_nplusn("SOxSS BC2", SS_SO_BC1, verbose=verbose)
    SS_SO_BC3_nplusn = SO.mate_nplusn("SOxSS BC3", SS_SO_BC2_nplusn, verbose=verbose)
    SS_SO_BC4_nplusn = SO.mate_nplusn("SOxSS BC4", SS_SO_BC3_nplusn, verbose=verbose)
    return [
        SS_SO_F1,
        SS_SO_BC1,
        SS_SO_BC2_nplusn,
        SS_SO_BC3_nplusn,
        SS_SO_BC4_nplusn,
    ]


def plot_summary(ax, samples: List[Genome]) -> GenomeSummary:
    """Plot the distribution of chromosome numbers given simulated samples.

    Args:
        ax (Axes): Matplotlib axes.
        samples (List[Genome]): Summarized genomes.

    Returns:
        GenomeSummary: Summary statistics of simulated genomes.
    """
    SO_data = []
    SS_data = []
    percent_SO_data = []
    percent_SS_data = []
    for s in samples:
        summary = s.summary
        SO_summary = [x for x in summary if x[0] == "SO"]
        group_unique = 0
        if SO_summary:
            _, group_unique, _, _ = SO_summary[0]
        SO_data.append(round(group_unique))
        SS_summary = [x for x in summary if x[0] == "SS"]
        group_unique = 0
        if SS_summary:
            _, group_unique, _, _ = SS_summary[0]
        SS_data.append(round(group_unique))
        total_tag, _, total_so_size, total_ss_size = summary[-1]
        assert total_tag == "Total"
        percent_SO_data.append(total_so_size * 100)
        percent_SS_data.append(total_ss_size * 100)
    # Avoid overlapping bars
    SS_counter, SO_counter = Counter(SS_data), Counter(SO_data)
    print(SS_counter, sum(SS_counter.values()))
    print(SO_counter, sum(SO_counter.values()))
    overlaps = SS_counter.keys() & SO_counter.keys()
    shift = 0.5  # used to offset bars a bit to avoid cluttering
    if overlaps:
        for overlap in overlaps:
            logger.debug("Modify bar offsets at %s due to SS and SO overlaps", overlap)
            SS_counter[overlap - shift] = SS_counter[overlap]
            del SS_counter[overlap]
            SO_counter[overlap + shift] = SO_counter[overlap]
            del SO_counter[overlap]

    def modify_range_end(d: dict, value: int):
        if value not in d:
            return
        # Has data at the range end, but no adjacent data points (i.e. isolated bar)
        if value in d and (value - 1 in d or value + 1 in d):
            return
        logger.debug("Modify bar offsets at %d due to end of range ends", value)
        d[value - shift if value else value + shift] = d[80]
        del d[value]

    modify_range_end(SS_counter, 0)
    modify_range_end(SS_counter, 80)
    modify_range_end(SO_counter, 0)
    modify_range_end(SO_counter, 80)

    x, y = zip(*sorted(SS_counter.items()))
    ax.bar(np.array(x), y, color=SsColor, ec=SsColor)
    x, y = zip(*sorted(SO_counter.items()))
    ax.bar(np.array(x), y, color=SoColor, ec=SoColor)
    ymax = len(samples) * 0.8
    ax.set_xlim(80, 0)
    ax.set_ylim(0, ymax)
    ax.set_yticks([])
    summary = GenomeSummary(SO_data, SS_data, percent_SO_data, percent_SS_data)

    # Write the stats summary within the plot
    summary_style = dict(size=8, va="center")
    SO_peak = SO_counter.most_common(1)[0][0]
    SS_peak = SS_counter.most_common(1)[0][0]
    SO_single = len(SO_counter) == 1
    SS_single = len(SS_counter) == 1

    for xpos, ypos, single, text, color, ha in zip(
        [SO_peak] * 4 + [SS_peak] * 4,
        ([ymax * 0.85] * 2 + [ymax * 0.65] * 2) * 2,
        [SO_single] * 4 + [SS_single] * 4,
        summary.SO_summary
        + summary.percent_SO_summary
        + summary.SS_summary
        + summary.percent_SS_summary,
        [SoColor] * 4 + [SsColor] * 4,
        ["right", "left"] * 4,
    ):
        # Offset some text to avoid overlapping
        if abs(SS_peak - SO_peak) < 20 and xpos == SO_peak:
            xpos += 12 if SO_peak > SS_peak else -12
        PAD = 1 if single else 0.25
        if ha == "left":
            xpos -= PAD
        else:
            xpos += PAD
        ax.text(xpos, ypos, markup(text), color=color, ha=ha, **summary_style)
    return summary


def write_chromosomes(genomes: List[Genome], filename: str):
    """Write simulated chromosomes to file

    Args:
        genomes (List[Genome]): List of simulated genomes.
        filename (str): File path to write to.
    """
    print(f"Write chromosomes to `{filename}`", file=sys.stderr)
    with open(filename, "w", encoding="utf-8") as fw:
        for genome in genomes:
            print(genome, file=fw)


def get_mode_title(mode: CrossMode) -> str:
    """
    Get the title of the mode.
    """
    if mode == CrossMode.nx2plusn:
        mode_title = r"$n_1\times2 + n_2$"
    elif mode == CrossMode.twoplusnFDR:
        mode_title = r"$2n + n$ (FDR)"
    elif mode == CrossMode.twoplusnSDR:
        mode_title = r"$n_1^*\times2 + n$ (SDR)"
    else:
        mode_title = "Unknown"
    return mode_title


def simulate(args):
    """
    %prog simulate [2n+n_FDR|2n+n_SDR|nx2+n]

    Run simulation on female restitution. There are several modes:
    - 2n+n_FDR: merger between a somatic and a germline
    - 2n+n_SDR: merger between a recombined germline and a germline (not yet supported)
    - nx2+n: merger between a doubled germline and a germline

    These modes would impact the sequence diversity in the progeny
    genome in F1, BCn ... the goal of this simulation, is thus to
    understand the mode and the spread of such diversity in the hybrid
    progenies.
    """
    sns.set_style("darkgrid")

    p = OptionParser(simulate.__doc__)
    p.add_argument(
        "--verbose",
        default=False,
        action="store_true",
        help="Verbose logging during simulation",
    )
    p.add_argument("-N", default=1000, type=int, help="Number of simulated samples")
    p.add_argument("--ss-ploidy", default=16, type=int, help="SS ploidy")
    opts, args, iopts = p.set_image_options(args, figsize="6x6")
    if len(args) != 1:
        sys.exit(not p.print_help())

    (mode,) = args
    SS_PLOIDY = opts.ss_ploidy
    SS_GENE_COUNT = SS_PLOIDY * HAPLOID_GENE_COUNT
    mode = CrossMode(mode)
    logger.info("Transmission: %s", mode)

    # Construct a composite figure with 6 tracks
    fig = plt.figure(1, (iopts.w, iopts.h))
    root = fig.add_axes([0, 0, 1, 1])
    rows = 5
    ypad = 0.05
    yinterval = (1 - 2 * ypad) / (rows + 1)
    yy = 1 - ypad
    xpad = 0.2
    xwidth = 0.7

    # Axes are vertically stacked, and share x-axis
    axes = []
    yy_positions = []  # Save yy positions so we can show details to the right later
    for idx in range(rows):
        yy_positions.append(yy)
        yy -= yinterval
        ax = fig.add_axes([xpad, yy, xwidth, yinterval * 0.85])
        if idx != rows - 1:
            plt.setp(ax.get_xticklabels(), visible=False)
        axes.append(ax)
    ax1, ax2, ax3, ax4, ax5 = axes

    # Prepare the simulated data
    # Simulate two parents
    SS = Genome("SS", "SS", SS_PLOIDY, SS_CHROM_COUNT, HAPLOID_GENE_COUNT)
    SO = Genome("SO", "SO", SO_PLOIDY, SO_CHROM_COUNT, HAPLOID_GENE_COUNT)

    verbose = opts.verbose
    N = opts.N
    with Pool(cpu_count()) as pool:
        results = pool.starmap(simulate_F1_to_BC4, [(SO, SS, mode, verbose)] * N)
    all_F1s, all_BC1s, all_BC2s, all_BC3s, all_BC4s = zip(*results)

    # Plotting
    plot_summary(ax1, all_F1s)
    plot_summary(ax2, all_BC1s)
    plot_summary(ax3, all_BC2s)
    plot_summary(ax4, all_BC3s)
    plot_summary(ax5, all_BC4s)

    # Show title to the left
    xx = xpad / 2
    for (title, subtitle), yy in zip(TITLES, yy_positions):
        if subtitle:
            yy -= 0.06
        else:
            yy -= 0.07
        root.text(
            xx,
            yy,
            title,
            color="darkslategray",
            ha="center",
            va="center",
            fontweight="semibold",
        )
        if subtitle:
            yy -= 0.02
            root.text(
                xx, yy, subtitle, color="lightslategray", ha="center", va="center"
            )

    axes[-1].set_xlabel("Number of unique chromosomes")
    adjust_spines(axes[-1], ["bottom"], outward=True)
    normalize_axes(root)

    # Title
    mode_title = get_mode_title(mode)
    root.text(0.5, 0.95, f"Transmission: {mode_title}", ha="center")

    savefig(f"{mode}.ss_ploidy_{SS_PLOIDY}.pdf", dpi=120)

    outdir = f"simulations_{mode}"
    mkdir(outdir)
    # Write chromosomes to disk
    for genomes, cross in zip(
        [all_F1s, all_BC1s, all_BC2s, all_BC3s, all_BC4s], CROSSES
    ):
        write_chromosomes(genomes, op.join(outdir, f"all_{cross}"))


def get_genome_wide_pct(summary: str) -> Dict[tuple, list]:
    """Collect genome-wide ungapped percent identity.
    Specifically, from file `SS_SR_SO.summary.txt`.

    Args:
        summary (str): File that contains per chromosome pct identity info,
        collected via `formats.blast.summary()`.

    Returns:
        Dict[tuple, list]: Genome pair to list of pct identities.
    """
    COLUMNS = "filename, identicals, qry_gapless, qry_gapless_pct, ref_gapless, ref_gapless_pct, qryspan, pct_qryspan, refspan, pct_refspan".split(
        ", "
    )
    df = pd.read_csv(summary, sep="\t", names=COLUMNS)
    data_by_genomes = defaultdict(list)
    for _, row in df.iterrows():
        filename = row["filename"]
        # e.g. SO_Chr01A.SO_Chr01B.1-1.blast
        chr1, chr2 = filename.split(".")[:2]
        genome1, chr1 = chr1.split("_")
        genome2, chr2 = chr2.split("_")
        chr1, chr2 = chr1[:5], chr2[:5]
        if (  # Special casing for SS certain chromosomes that are non-collinear with SO/SR
            genome1 != "SS"
            and genome2 == "SS"
            and chr2 not in ("Chr01", "Chr03", "Chr04")
        ):
            continue
        qry_gapless_pct, ref_gapless_pct = (
            row["qry_gapless_pct"],
            row["ref_gapless_pct"],
        )
        data_by_genomes[(genome1, genome2)] += [qry_gapless_pct, ref_gapless_pct]
    return data_by_genomes


def get_anchors_pct(filename: str, min_pct: int = 94) -> list:
    """Collect CDS-wide ungapped percent identity.

    Args:
        filename (str): Input file name, which is a LAST file.

    Returns:
        list: List of pct identities from this LAST file.
    """
    blast = Blast(filename)
    pct = []
    for c in blast:
        if c.pctid < min_pct:
            continue
        identicals = c.hitlen - c.nmismatch - c.ngaps
        qstart, qstop = c.qstart, c.qstop
        sstart, sstop = c.sstart, c.sstop
        qrycovered = qstop - qstart + 1
        refcovered = sstop - sstart + 1
        pct.append(identicals * 100 / qrycovered)
        pct.append(identicals * 100 / refcovered)
    return pct


def divergence(args):
    """
    %prog divergence SS_SR_SO.summary.txt

    Plot divergence between and within SS/SR/SO genomes.
    """
    sns.set_style("white")

    p = OptionParser(divergence.__doc__)
    p.add_argument("--title", default="Gapless", help="Plot title")
    p.add_argument(
        "--xmin",
        default=94,
        type=int,
        help="Minimum percent identity in the histogram",
    )
    opts, args, iopts = p.set_image_options(args, figsize="8x8")
    if len(args) != 1:
        sys.exit(not p.print_help())

    (summary,) = args
    data_by_genomes = get_genome_wide_pct(summary)
    # Print summary statistics
    print("Genome-wide ungapped percent identity:")
    for (genome1, genome2), pct in sorted(data_by_genomes.items()):
        print(genome1, genome2, np.mean(pct), np.std(pct))

    # Plotting genome-wide divergence
    fig = plt.figure(figsize=(iopts.w, iopts.h))
    root = fig.add_axes([0, 0, 1, 1])
    SPECIES_CONFIG = {
        "SS": {"label": "S. spontaneum", "pos": (0.5, 0.67)},
        "SR": {"label": "S. robustum", "pos": (0.2, 0.3)},
        "SO": {"label": "S. officinarum", "pos": (0.8, 0.3)},
    }
    # Get median for each genome pair
    medians = {}
    for g1, g2 in product(SPECIES_CONFIG.keys(), repeat=2):
        g1, g2 = sorted((g1, g2))
        d = data_by_genomes[(g1, g2)]
        medians[(g1, g2)] = np.median(d)
    for g, config in SPECIES_CONFIG.items():
        x, y = config["pos"]
        text = f'*{config["label"]}*' + f"\n{medians[(g, g)]:.1f} %"
        text = markup(text)
        root.text(x, y, text, color="darkslategray", ha="center", va="center")

    # Connect lines
    PAD, YPAD = 0.09, 0.045
    for g1, g2 in combinations(SPECIES_CONFIG.keys(), 2):
        g1, g2 = sorted((g1, g2))
        x1, y1 = SPECIES_CONFIG[g1]["pos"]
        x2, y2 = SPECIES_CONFIG[g2]["pos"]
        x1, x2 = (x1 + PAD, x2 - PAD) if x1 < x2 else (x1 - PAD, x2 + PAD)
        if y1 != y2:
            y1, y2 = (y1 + YPAD, y2 - YPAD) if y1 < y2 else (y1 - YPAD, y2 + YPAD)
        xmid, ymid = (x1 + x2) / 2, (y1 + y2) / 2
        text = f"{medians[(g1, g2)]:.1f} %"
        text = markup(text)
        root.text(xmid, ymid, text, ha="center", va="center", backgroundcolor="w")
        root.plot([x1, x2], [y1, y2], "-", lw=4, color="darkslategray")

    # Pct identity histograms
    PCT_CONFIG = {
        ("SS", "SS"): {"pos": (0.5, 0.82)},
        ("SR", "SR"): {"pos": (0.1, 0.2)},
        ("SO", "SO"): {"pos": (0.9, 0.2)},
        ("SR", "SS"): {"pos": (0.3 - PAD, 0.55)},
        ("SO", "SS"): {"pos": (0.7 + PAD, 0.55)},
        ("SO", "SR"): {"pos": (0.5, 0.18)},
    }
    HIST_WIDTH = 0.15
    xmin = opts.xmin
    for genome_pair, config in PCT_CONFIG.items():
        x, y = config["pos"]
        ax = fig.add_axes(
            [x - HIST_WIDTH / 2, y - HIST_WIDTH / 2, HIST_WIDTH, HIST_WIDTH]
        )
        d = data_by_genomes[genome_pair]
        binwidth = (100 - xmin) / 20
        sns.histplot(d, ax=ax, binwidth=binwidth, kde=False)
        ax.set_xlim(xmin, 100)
        ax.get_yaxis().set_visible(False)
        ax.set_xticks([xmin, 100])
        adjust_spines(ax, ["bottom"], outward=True)
        ax.spines["bottom"].set_color("lightslategray")

    title = opts.title
    italic_title = markup(f"*{title}*")
    root.text(
        0.5,
        0.95,
        f"{italic_title} identities between and within SS/SR/SO genomes",
        size=14,
        ha="center",
        va="center",
    )
    normalize_axes(root)
    image_name = f"SO_SR_SS.{title}." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


def plot_genome(
    ax,
    x: float,
    y: float,
    genome: Genome,
    haplotype_gene_height: Dict[str, float],
    haplotype_colors: Dict[str, str],
    chrom_width: float = 0.012,
    gap_width: float = 0.008,
    pair: bool = False,
):
    """
    Plot the genome in the axes, centered around (x, y).
    """
    target = "chr01"  # Arbitrary target chromosome
    ploidy = genome.ploidy(target)
    total_width = ploidy * (chrom_width + gap_width) - gap_width
    xx = x - total_width / 2
    if pair:
        genome.pair_chromosomes(inplace=True)
        print(
            genome.name, ":", [str(x) for x in genome.chromosomes if x.chrom == target]
        )
    for chrom in genome.chromosomes:
        if chrom.chrom != target:
            continue
        gene_height = haplotype_gene_height[chrom.subgenome]
        height = gene_height * len(chrom)
        ChromosomePlot(ax, xx, y, y - height, ec="lightslategray")
        yy = y
        subgenome = chrom.subgenome
        for haplotype, genes in groupby(chrom.genes, key=lambda x: x.haplotype):
            genes = list(genes)
            g1, g2 = genes[0].idx - 1, genes[-1].idx
            patch_height = gene_height * (g2 - g1)
            color = haplotype_colors[subgenome, haplotype]
            ax.add_patch(
                Rectangle(
                    (xx - chrom_width / 2, yy - patch_height),
                    chrom_width,
                    patch_height,
                    fc=color,
                    lw=0,
                )
            )
            yy -= patch_height
        xx += chrom_width + gap_width


def truncate_SS_chromosome(genome: Genome, cross: str, ss_ploidy: int) -> Genome:
    """
    Truncate the last SS chromosome to illustrate half/quarter ploidy.
    """
    subgenome, chrom = "SS", "chr01"
    if ss_ploidy == 8:
        if cross == "BC3":
            genome.truncate_chromosome(subgenome, chrom, 0.5)
        elif cross == "BC4":
            genome.truncate_chromosome(subgenome, chrom, 0.25)
    elif ss_ploidy == 16:
        if cross == "BC4":
            genome.truncate_chromosome(subgenome, chrom, 0.5)
    else:
        raise ValueError("Unsupported SS ploidy for custom mode")
    return genome


def chromosome(args):
    """
    %prog chromosome [2n+n_FDR|2n+n_SDR|nx2+n]
    """
    p = OptionParser(simulate.__doc__)
    p.add_argument("-k", default=0, type=int, help="Plot k-th simulated genomes")
    p.add_argument("--ss-ploidy", default=16, type=int, help="SS ploidy")
    p.add_argument(
        "--custom",
        default=False,
        action="store_true",
        help="Custom mode to truncate BC3/4 SS chromosome",
    )
    opts, args, iopts = p.set_image_options(args, figsize="6x6")
    if len(args) != 1:
        sys.exit(not p.print_help())

    (mode,) = args
    SS_PLOIDY = opts.ss_ploidy
    SS_GENE_COUNT = SS_PLOIDY * HAPLOID_GENE_COUNT
    mode = CrossMode(mode)
    logger.info("Transmission: %s", mode)
    if opts.custom:
        logger.info("Custom mode: truncate SS chromosome")

    # Construct a composite figure with 6 tracks
    fig = plt.figure(1, (iopts.w, iopts.h))
    root = fig.add_axes([0, 0, 1, 1])
    rows = 5
    ypad = 0.05
    yinterval = (1 - 2 * ypad) / (rows + 1)
    yy = 1 - ypad
    xpad = 0.2

    SS = Genome("SS", "SS", SS_PLOIDY, SS_CHROM_COUNT, HAPLOID_GENE_COUNT)
    SO = Genome("SO", "SO", SO_PLOIDY, SO_CHROM_COUNT, HAPLOID_GENE_COUNT)

    indir = f"simulations_{mode}"
    genomes = []
    for cross in CROSSES:
        filename = op.join(indir, f"all_{cross}")
        with open(filename, encoding="utf-8") as fp:
            for i, row in enumerate(fp):
                if i != opts.k:
                    continue
                genome = Genome.from_str(row)
                if opts.custom:
                    truncate_SS_chromosome(genome, cross, opts.ss_ploidy)
                break
        genomes.append((cross, genome))

    yy_positions = []  # Save yy positions so we can show details to the right later
    for idx in range(rows + 1):
        yy_positions.append(yy)
        yy -= yinterval
        if idx != rows:
            root.plot([0, 1], [yy, yy], lw=0.5, color="gray")

    # Show title to the left
    xx = xpad / 2
    titles = (("Progenitors", None),) + TITLES
    for (title, _), yy in zip(titles, yy_positions):
        yy -= 0.07
        root.text(
            xx,
            yy,
            markup(title),
            color="darkslategray",
            ha="center",
            va="center",
            fontweight="semibold",
        )

    SO_colors = get_shades(SoColor, SO_PLOIDY)
    SS_colors = get_shades(SsColor, SS_PLOIDY)
    SO_haplotypes = [("SO", chr(ord("a") + i)) for i in range(SO_PLOIDY)]
    SS_haplotypes = [("SS", chr(ord("a") + i)) for i in range(SS_PLOIDY)]
    SO_haplotype_colors = dict(zip(SO_haplotypes, SO_colors))
    SS_haplotype_colors = dict(zip(SS_haplotypes, SS_colors))
    haplotype_colors = {**SO_haplotype_colors, **SS_haplotype_colors}

    # Plotting
    chrom_height = 0.1
    SO_GENE_HEIGHT = chrom_height / (HAPLOID_GENE_COUNT // SO_CHROM_COUNT)
    SS_GENE_HEIGHT = chrom_height / (HAPLOID_GENE_COUNT // SS_CHROM_COUNT)
    haplotype_gene_height = {"SO": SO_GENE_HEIGHT, "SS": SS_GENE_HEIGHT}
    yy = 0.92
    plot_genome(root, 0.35, yy, SO, haplotype_gene_height, haplotype_colors)
    plot_genome(
        root,
        0.75 if SS_PLOIDY == 16 else 0.66,
        yy,
        SS,
        haplotype_gene_height,
        haplotype_colors,
    )
    # Plot big cross sign
    root.text(
        0.5, yy - chrom_height / 2, r"$\times$", ha="center", va="center", fontsize=36
    )
    # Genome labels
    root.text(
        0.215,
        yy - chrom_height / 2,
        markup(f"*So*\n({SO_PLOIDY}x)"),
        ha="center",
        va="center",
        color=SoColor,
    )
    root.text(
        0.945 if SS_PLOIDY == 16 else 1 - 0.215,
        yy - chrom_height / 2,
        markup(f"*Ss*\n({SS_PLOIDY}x)"),
        ha="center",
        va="center",
        color=SsColor,
    )

    for _, genome in genomes:
        yy -= yinterval
        plot_genome(
            root, 0.5, yy, genome, haplotype_gene_height, haplotype_colors, pair=True
        )

    # Title
    mode_title = get_mode_title(mode)
    root.text(0.5, 0.95, f"Transmission: {mode_title}", ha="center")
    normalize_axes(root)

    savefig(f"{mode}.chromosome.ss_ploidy_{SS_PLOIDY}.pdf", dpi=120)


def main():

    actions = (
        ("simulate", "Run simulation on female restitution"),
        # Plot the simulated chromosomes
        ("chromosome", "Plot the chromosomes of the simulated genomes"),
        # Plotting scripts to illustrate divergence between and within genomes
        ("divergence", "Plot divergence between and within SS/SR/SO genomes"),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Read-based phasing.
"""
import sys

try:
    import vcf
except ImportError:
    pass
import pysam

from ..apps.base import ActionDispatcher, OptionParser, logger


class CPRA:
    def __init__(self, vcf_record):
        r = vcf_record
        self.chr = r.CHROM
        self.pos = r.POS
        self.ref = r.REF
        self.alt = r.ALT
        self.alleles = [self.ref] + self.alt

    @property
    def is_valid(self):
        """Only retain SNPs or single indels, and are bi-allelic"""
        return len(self.ref) == 1 and len(self.alt) == 1 and len(self.alt[0]) == 1

    def __str__(self):
        return "_".join(str(x) for x in (self.chr, self.pos, self.ref, self.alt[0]))

    __repr__ = __str__


def main():

    actions = (
        ("prepare", "convert vcf and bam to variant list"),
        ("counts", "collect allele counts from RO/AO fields"),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def counts(args):
    """
    %prog counts vcffile

    Collect allele counts from RO and AO fields.
    """
    p = OptionParser(counts.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (vcffile,) = args
    vcf_reader = vcf.Reader(open(vcffile))
    for r in vcf_reader:
        v = CPRA(r)
        if not v.is_valid:
            continue
        for sample in r.samples:
            ro = sample["RO"]
            ao = sample["AO"]
            print("\t".join(str(x) for x in (v, ro, ao)))


def prepare(args):
    """
    %prog prepare vcffile bamfile

    Convert vcf and bam to variant list. Inputs are:
    - vcffile: contains the positions of variants
    - bamfile: contains the reads that hold the variants

    Outputs:
    - reads_to_phase: phasing for each read
    - variants_to_phase: in format of phased vcf
    """
    p = OptionParser(prepare.__doc__)
    p.add_argument("--accuracy", default=0.85, help="Sequencing per-base accuracy")
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    vcffile, bamfile = args
    right = "{:.2f}".format(opts.accuracy)
    wrong = "{:.2f}".format(1 - opts.accuracy)
    vcf_reader = vcf.Reader(open(vcffile))
    variants = []
    for r in vcf_reader:
        v = CPRA(r)
        if not v.is_valid:
            continue
        variants.append(v)

    logger.debug(
        "A total of %d bi-allelic SNVs imported from `%s`", len(variants), vcffile
    )

    bamfile = pysam.AlignmentFile(bamfile, "rb")
    for v in variants:
        pos = v.pos - 1
        for column in bamfile.pileup(v.chr, pos, pos + 1, truncate=True):
            for read in column.pileups:
                query_position = read.query_position
                if query_position is None:
                    continue
                read_name = read.alignment.query_name
                query_base = read.alignment.query_sequence[query_position]
                a, b = v.alleles
                if query_base == a:
                    other_base = b
                elif query_base == b:
                    other_base = a
                else:
                    continue
                print(
                    " ".join(
                        str(x)
                        for x in (v, read_name, query_base, right, other_base, wrong)
                    )
                )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Manipulate PDF files, using PyPDF library.
"""
import shutil
import sys

from natsort import natsorted
from pypdf import PdfWriter
from pypdf.pagerange import PageRange, parse_filename_page_ranges  # stable home

from ..apps.base import ActionDispatcher, OptionParser, cleanup, logger
from .base import must_open

PAGE_RANGE_HELP = PageRange.__init__.__doc__


def main():

    actions = (("cat", "concatenate pages from pdf files into a single pdf file"),)
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def cat(args):
    """
    %prog cat *.pdf -o output.pdf

    Concatenate pages from pdf files into a single pdf file.

    Page ranges refer to the previously-named file.
    A file not followed by a page range means all the pages of the file.

    PAGE RANGES are like Python slices.
            {page_range_help}
    EXAMPLES
        pdfcat -o output.pdf head.pdf content.pdf :6 7: tail.pdf -1
            Concatenate all of head.pdf, all but page seven of content.pdf,
            and the last page of tail.pdf, producing output.pdf.

        pdfcat chapter*.pdf >book.pdf
            You can specify the output file by redirection.

        pdfcat chapter?.pdf chapter10.pdf >book.pdf
            In case you don't want chapter 10 before chapter 2.
    """
    p = OptionParser(cat.__doc__.format(page_range_help=PAGE_RANGE_HELP))
    p.add_argument(
        "--nosort", default=False, action="store_true", help="Do not sort file names"
    )
    p.add_argument(
        "--cleanup",
        default=False,
        action="store_true",
        help="Remove individual pdfs after merging",
    )
    p.set_outfile()
    opts, args = p.parse_args(args)

    if len(args) < 1:
        sys.exit(not p.print_help())

    outfile = opts.outfile
    if outfile in args:
        args.remove(outfile)

    should_sort = not opts.nosort
    if not all(x.endswith(".pdf") for x in args):
        should_sort = False
        logger.debug("Not sorting filenames because non-pdf args")

    if should_sort:
        args = natsorted(args)

    filename_page_ranges = parse_filename_page_ranges(args)
    nfiles = len(filename_page_ranges)

    # Fast-path: if exactly one input and range == ":" (all), do a simple copy
    if nfiles == 1:
        in_file, pr = filename_page_ranges[0]
        sl = getattr(pr, "to_slice", lambda: slice(None))()
        if sl.start is None and sl.stop is None and sl.step is None:
            logger.debug("Single input with full range detected; copying bytes")
            try:
                shutil.copyfile(in_file, outfile)
                logger.info("Copied `%s` to `%s`", in_file, outfile)
            except Exception as e:
                logger.error("Error while copying %s: %s", in_file, e)
                sys.exit(1)
            if opts.cleanup:
                logger.debug("Cleaning up 1 file")
                cleanup([in_file])
            return

    writer = PdfWriter()
    try:
        for filename, page_range in filename_page_ranges:
            logger.debug("%s: %s", filename, page_range)
            # `append` accepts PageRange objects directly (preferred over slices)
            writer.append(fileobj=filename, pages=page_range)

        with must_open(outfile, "wb") as fw:
            writer.write(fw)
        writer.close()
        logger.info("Extracted %d files into `%s`", nfiles, outfile)
    except Exception as e:
        logger.error("Error while reading %s: %s", filename, e)  # noqa: F821
        sys.exit(1)

    if opts.cleanup:
        logger.debug("Cleaning up %d files", nfiles)
        cleanup(args)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Utilities for submitting PASA jobs and processing PASA results.
"""
import os
import os.path as op
import sys

from ..apps.base import ActionDispatcher, OptionParser, logger, sh, symlink, which
from ..formats.base import FileMerger, must_open, write_file

alignAssembly_conf = """
# MySQL settings
MYSQLDB={0}

#script validate_alignments_in_db.dbi
validate_alignments_in_db.dbi:--MIN_PERCENT_ALIGNED={1}
validate_alignments_in_db.dbi:--MIN_AVG_PER_ID={2}
validate_alignments_in_db.dbi:--NUM_BP_PERFECT_SPLICE_BOUNDARY={3}

#script subcluster_builder.dbi
subcluster_builder.dbi:-m=50
"""

annotCompare_conf = """
# MySQL settings
MYSQLDB={0}

#script cDNA_annotation_comparer.dbi
cDNA_annotation_comparer.dbi:--MIN_PERCENT_OVERLAP={1}
cDNA_annotation_comparer.dbi:--MIN_PERCENT_PROT_CODING={2}
cDNA_annotation_comparer.dbi:--MIN_PERID_PROT_COMPARE={3}
cDNA_annotation_comparer.dbi:--MIN_PERCENT_LENGTH_FL_COMPARE={4}
cDNA_annotation_comparer.dbi:--MIN_PERCENT_LENGTH_NONFL_COMPARE={5}
cDNA_annotation_comparer.dbi:--MIN_FL_ORF_SIZE={6}
cDNA_annotation_comparer.dbi:--MIN_PERCENT_ALIGN_LENGTH={7}
cDNA_annotation_comparer.dbi:--MIN_PERCENT_OVERLAP_GENE_REPLACE={8}
cDNA_annotation_comparer.dbi:--STOMP_HIGH_PERCENTAGE_OVERLAPPING_GENE={9}
cDNA_annotation_comparer.dbi:--TRUST_FL_STATUS={10}
cDNA_annotation_comparer.dbi:--MAX_UTR_EXONS={11}
"""

annotation = "annotation.gff3"
tdn, flaccs = "tdn.accs", "FL_accs.txt"
tfasta, gfasta = "transcripts.fasta", "genome.fasta"
aaconf, acconf = "alignAssembly.conf", "annotCompare.conf"
ALLOWED_ALIGNERS = ("blat", "gmap")


def main():

    actions = (
        ("assemble", "run pasa alignment assembly pipeline"),
        ("compare", "run pasa annotation comparison pipeline"),
        ("longest", "label longest transcript per gene as full-length"),
        (
            "consolidate",
            "generate consolidated annotation set from 2 or more annot compare results",
        ),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def assemble(args):
    """
    %prog assemble pasa_db_name genome.fasta transcripts-dn.fasta [transcript-gg.fasta]

    Run the PASA alignment assembly pipeline

    If two transcript fasta files (Trinity denovo and genome guided) are provided
    and the `--compreh` param is enabled, the PASA Comprehensive Transcriptome DB
    protocol is followed <http://pasa.sourceforge.net/#A_ComprehensiveTranscriptome>

    Using the `--prepare` option creates a shell script with the run commands without
    executing the pipeline
    """
    p = OptionParser(assemble.__doc__)
    p.set_pasa_opts()
    p.add_argument(
        "--prepare",
        default=False,
        action="store_true",
        help="Prepare PASA run script with commands",
    )
    p.set_grid()
    p.set_grid_opts()
    opts, args = p.parse_args(args)

    if len(args) not in (3, 4):
        sys.exit(not p.print_help())

    (
        pasa_db,
        genome,
        dnfasta,
    ) = args[:3]
    ggfasta = args[3] if len(args) == 4 else None

    PASA_HOME = opts.pasa_home
    if not op.isdir(PASA_HOME):
        logger.error("PASA_HOME={0} directory does not exist".format(PASA_HOME))
        sys.exit()

    aligners = opts.aligners.split(",")
    for aligner in aligners:
        if aligner not in ALLOWED_ALIGNERS:
            logger.error("Error: Unknown aligner `{0}`".format(aligner))
            logger.error(
                "Can be any of {0}, ".format("|".join(ALLOWED_ALIGNERS))
                + "combine multiple aligners in list separated by comma"
            )
            sys.exit()

    clean = opts.clean
    seqclean = op.join(opts.tgi_home, "seqclean")

    accn_extract = which(op.join(PASA_HOME, "misc_utilities", "accession_extractor.pl"))
    launch_pasa = which(op.join(PASA_HOME, "scripts", "Launch_PASA_pipeline.pl"))
    build_compreh_trans = which(
        op.join(PASA_HOME, "scripts", "build_comprehensive_transcriptome.dbi")
    )

    fl_accs = opts.fl_accs
    cpus = opts.cpus
    grid = opts.grid
    prepare, runfile = opts.prepare, "run.sh"
    pctcov, pctid = opts.pctcov, opts.pctid
    compreh_pctid = opts.compreh_pctid
    compreh_pctcov, bpsplice = opts.compreh_pctcov, opts.bpsplice

    cmds = []

    # set PASAHOME env variable if preparing shell script
    if prepare:
        env_cmd = 'export PASAHOME="{0}"'.format(PASA_HOME)
        cmds.append(env_cmd)

    if ggfasta:
        transcripts = FileMerger([dnfasta, ggfasta], tfasta).merge()
        accn_extract_cmd = "cat {0} | {1} > {2}".format(dnfasta, accn_extract, tdn)
        cmds.append(accn_extract_cmd)
        if not prepare:
            sh(accn_extract_cmd)
    else:
        symlink(dnfasta, tfasta)
        transcripts = tfasta

    if opts.grid and not opts.threaded:
        opts.threaded = opts.cpus

    prjobid = None
    if clean:
        ccpus = 16 if cpus >= 16 else cpus
        cleancmd = "{0} {1} -c {2} -l 60".format(seqclean, transcripts, ccpus)
        if prepare:
            cmds.append(cleancmd)
        else:
            prjobid = sh(cleancmd, grid=grid, grid_opts=opts)

    aafw = must_open(aaconf, "w")
    print(
        alignAssembly_conf.format("{0}_pasa".format(pasa_db), pctcov, pctid, bpsplice),
        file=aafw,
    )
    aafw.close()

    symlink(genome, gfasta)

    aacmd = "{0} -c {1} -C -R -g {2}".format(launch_pasa, aaconf, gfasta)
    aacmd += (
        " -t {0}.clean -T -u {0}".format(transcripts)
        if clean
        else " -t {0}".format(transcripts)
    )
    if fl_accs:
        symlink(fl_accs, flaccs)
        aacmd += " -f {0}".format(flaccs)
    if ggfasta:
        aacmd += " --TDN {0}".format(tdn)
    aacmd += " --ALIGNERS {0} -I {1} --CPU {2}".format(
        ",".join(aligners), opts.intron, cpus
    )

    if prepare:
        cmds.append(aacmd)
    else:
        opts.hold_jid = prjobid
        prjobid = sh(aacmd, grid=grid, grid_opts=opts)

    if opts.compreh and ggfasta:
        comprehcmd = "{0} -c {1} -t {2}".format(
            build_compreh_trans, aaconf, transcripts
        )
        comprehcmd += " --min_per_ID {0} --min_per_aligned {1}".format(
            compreh_pctid, compreh_pctcov
        )

        if prepare:
            cmds.append(comprehcmd)
        else:
            opts.hold_jid = prjobid
            prjobid = sh(comprehcmd, grid=grid, grid_opts=opts)

    if prepare:
        write_file(runfile, "\n".join(cmds))  # initialize run script


def compare(args):
    """
    %prog compare pasa_db_name [--annots_gff3=annotation.gff3]

    Run the PASA annotation comparison pipeline

    This assumes that PASA alignment assembly has alredy been completed and
    run directory contains `genome.fasta` and `transcript.fasta` files.

    If `--annots_gff3` is specified, the PASA database is loaded with the annotations
    first before starting annotation comparison. Otherwise, it uses previously
    loaded annotation data.

    Using the `--prepare` option creates a shell script with the run commands without
    executing the pipeline
    """
    p = OptionParser(compare.__doc__)
    p.set_pasa_opts(action="compare")
    p.add_argument(
        "--prepare",
        default=False,
        action="store_true",
        help="Prepare PASA run script with commands",
    )
    p.set_grid()
    p.set_grid_opts()
    opts, args = p.parse_args(args)

    if len(args) < 1:
        sys.exit(not p.print_help())

    (pasa_db,) = args

    PASA_HOME = opts.pasa_home
    if not op.isdir(PASA_HOME):
        logger.error("PASA_HOME={0} directory does not exist".format(PASA_HOME))
        sys.exit()

    launch_pasa = which(op.join(PASA_HOME, "scripts", "Launch_PASA_pipeline.pl"))

    annots_gff3 = opts.annots_gff3
    grid = opts.grid
    prepare, runfile = opts.prepare, "run.sh"

    os.chdir(pasa_db)

    if prepare:
        write_file(runfile, "", append=True, skipcheck=True)  # initialize run script

    acfw = must_open(acconf, "w")
    print(
        annotCompare_conf.format(
            "{0}_pasa".format(pasa_db),
            opts.pctovl,
            opts.pct_coding,
            opts.pctid_prot,
            opts.pctlen_FL,
            opts.pctlen_nonFL,
            opts.orf_size,
            opts.pct_aln,
            opts.pctovl_gene,
            opts.stompovl,
            opts.trust_FL,
            opts.utr_exons,
        ),
        file=acfw,
    )
    acfw.close()

    if not op.exists(gfasta):
        sys.exit("Genome fasta file `{0}` does not exist".format(gfasta))

    transcripts = tfasta
    if not op.exists(transcripts):
        sys.exit("Transcript fasta file `{0}` does not exist".format(transcripts))

    if op.exists("{0}.clean".format(transcripts)):
        transcripts = "{0}.clean".format(transcripts)

    accmd = "{0} -c {1} -A -g {2} -t {3} --GENETIC_CODE {4}".format(
        launch_pasa, acconf, gfasta, transcripts, opts.genetic_code
    )

    if annots_gff3:
        if not op.exists(annots_gff3):
            sys.exit("Annotation gff3 file `{0}` does not exist".format(annots_gff3))
        symlink(annots_gff3, annotation)
        accmd += " -L --annots_gff3 {0}".format(annotation)

    if prepare:
        write_file(runfile, accmd, append=True)
    else:
        sh(accmd, grid=grid, grid_opts=opts)


def longest(args):
    """
    %prog longest pasa.fasta output.subclusters.out

    Find the longest PASA assembly and label it as full-length. Also removes
    transcripts shorter than half the length of the longest, or shorter than
    200bp. The assemblies for the same locus is found in
    `output.subclusters.out`. In particular the lines that look like:

    sub-cluster: asmbl_25 asmbl_26 asmbl_27
    """
    from jcvi.formats.fasta import Fasta, SeqIO
    from jcvi.formats.sizes import Sizes

    p = OptionParser(longest.__doc__)
    p.add_argument(
        "--prefix",
        default="pasa",
        help="Replace asmbl_ with prefix",
    )
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    fastafile, subclusters = args
    prefix = fastafile.rsplit(".", 1)[0]

    idsfile = prefix + ".fl.ids"
    fw = open(idsfile, "w")
    sizes = Sizes(fastafile).mapping

    name_convert = lambda x: x.replace("asmbl", opts.prefix)

    keep = set()  # List of IDs to write
    fp = open(subclusters)
    nrecs = 0
    for row in fp:
        if not row.startswith("sub-cluster:"):
            continue
        asmbls = row.split()[1:]
        longest_asmbl = max(asmbls, key=lambda x: sizes[x])
        longest_size = sizes[longest_asmbl]
        print(name_convert(longest_asmbl), file=fw)
        nrecs += 1
        cutoff = max(longest_size / 2, 200)
        keep.update(set(x for x in asmbls if sizes[x] >= cutoff))

    fw.close()
    logger.debug("{0} fl-cDNA records written to `{1}`.".format(nrecs, idsfile))

    f = Fasta(fastafile, lazy=True)
    newfastafile = prefix + ".clean.fasta"
    fw = open(newfastafile, "w")
    nrecs = 0
    for name, rec in f.iteritems_ordered():
        if name not in keep:
            continue

        rec.id = name_convert(name)
        rec.description = ""
        SeqIO.write([rec], fw, "fasta")
        nrecs += 1

    fw.close()
    logger.debug("{0} valid records written to `{1}`.".format(nrecs, newfastafile))


def consolidate(args):
    """
    %prog consolidate gffile1 gffile2 ... > consolidated.out

    Given 2 or more gff files generated by pasa annotation comparison,
    iterate through each locus (shared locus name or overlapping CDS)
    and identify same/different isoforms (shared splicing structure)
    across the input datasets.

    If `slop` is enabled, consolidation will collapse any variation
    in terminal UTR lengths, keeping the longest as representative.
    """
    from itertools import combinations, product

    from jcvi.formats.base import longest_unique_prefix
    from jcvi.formats.gff import make_index, match_subfeats
    from jcvi.utils.cbook import AutoVivification
    from jcvi.utils.grouper import Grouper

    supported_modes = ["name", "coords"]
    p = OptionParser(consolidate.__doc__)
    p.add_argument(
        "--slop",
        default=False,
        action="store_true",
        help="allow minor variation in terminal 5'/3' UTR start/stop position",
    )
    p.add_argument(
        "--inferUTR",
        default=False,
        action="store_true",
        help="infer presence of UTRs from exon coordinates",
    )
    p.add_argument(
        "--mode",
        default="name",
        choices=supported_modes,
        help="method used to determine overlapping loci",
    )
    p.add_argument(
        "--summary",
        default=False,
        action="store_true",
        help="Generate summary table of consolidation process",
    )
    p.add_argument(
        "--clusters",
        default=False,
        action="store_true",
        help="Generate table of cluster members after consolidation",
    )
    p.set_outfile()

    opts, args = p.parse_args(args)
    slop = opts.slop
    inferUTR = opts.inferUTR
    mode = opts.mode

    if len(args) < 2:
        sys.exit(not p.print_help())

    gffdbx = {}
    for gffile in args:
        dbn = longest_unique_prefix(gffile, args)
        gffdbx[dbn] = make_index(gffile)

    loci = Grouper()
    for dbn in gffdbx:
        odbns = [odbn for odbn in gffdbx if dbn != odbn]
        for gene in gffdbx[dbn].features_of_type("gene", order_by=("seqid", "start")):
            if mode == "name":
                loci.join(gene.id, (gene.id, dbn))
            else:
                if (gene.id, dbn) not in loci:
                    loci.join((gene.id, dbn))
                    gene_cds = list(
                        gffdbx[dbn].children(gene, featuretype="CDS", order_by="start")
                    )
                    gene_cds_start, gene_cds_stop = gene_cds[0].start, gene_cds[-1].stop
                    for odbn in odbns:
                        for ogene_cds in gffdbx[odbn].region(
                            seqid=gene.seqid,
                            start=gene_cds_start,
                            end=gene_cds_stop,
                            strand=gene.strand,
                            featuretype="CDS",
                        ):
                            for ogene in gffdbx[odbn].parents(
                                ogene_cds, featuretype="gene"
                            ):
                                loci.join((gene.id, dbn), (ogene.id, odbn))

    gfeats = {}
    mrna = AutoVivification()
    for i, locus in enumerate(loci):
        gene = "gene_{0:0{pad}}".format(i, pad=6) if mode == "coords" else None

        for elem in locus:
            if type(elem) == tuple:
                _gene, dbn = elem
                if gene is None:
                    gene = _gene

                g = gffdbx[dbn][_gene]
                if gene not in gfeats:
                    gfeats[gene] = g
                    gfeats[gene].attributes["ID"] = [gene]
                else:
                    if g.start < gfeats[gene].start:
                        gfeats[gene].start = g.start
                    if g.stop > gfeats[gene].stop:
                        gfeats[gene].stop = g.stop

                c = list(
                    gffdbx[dbn].children(_gene, featuretype="mRNA", order_by="start")
                )
                if len(c) > 0:
                    mrna[gene][dbn] = c

    fw = must_open(opts.outfile, "w")
    print("##gff-version	3", file=fw)
    seen = {}
    if opts.summary:
        summaryfile = "{0}.summary.txt".format(opts.outfile.rsplit(".")[0])
        sfw = must_open(summaryfile, "w")
        summary = ["id"]
        summary.extend(gffdbx.keys())
        print("\t".join(str(x) for x in summary), file=sfw)
    if opts.clusters:
        clustersfile = "{0}.clusters.txt".format(opts.outfile.rsplit(".")[0])
        cfw = must_open(clustersfile, "w")
        clusters = ["id", "dbns", "members", "trlens"]
        print("\t".join(str(x) for x in clusters), file=cfw)
    for gene in mrna:
        g = Grouper()
        dbns = list(combinations(mrna[gene], 2))
        if len(dbns) > 0:
            for dbn1, dbn2 in dbns:
                dbx1, dbx2 = gffdbx[dbn1], gffdbx[dbn2]
                for mrna1, mrna2 in product(mrna[gene][dbn1], mrna[gene][dbn2]):
                    mrna1s, mrna2s = (
                        mrna1.stop - mrna1.start + 1,
                        mrna2.stop - mrna2.start + 1,
                    )
                    g.join((dbn1, mrna1.id, mrna1s))
                    g.join((dbn2, mrna2.id, mrna2s))

                    if match_subfeats(mrna1, mrna2, dbx1, dbx2, featuretype="CDS"):
                        res = []
                        ftypes = (
                            ["exon"]
                            if inferUTR
                            else ["five_prime_UTR", "three_prime_UTR"]
                        )
                        for ftype in ftypes:
                            res.append(
                                match_subfeats(
                                    mrna1,
                                    mrna2,
                                    dbx1,
                                    dbx2,
                                    featuretype=ftype,
                                    slop=slop,
                                )
                            )

                        if all(res):
                            g.join((dbn1, mrna1.id, mrna1s), (dbn2, mrna2.id, mrna2s))
        else:
            for dbn1 in mrna[gene]:
                for mrna1 in mrna[gene][dbn1]:
                    g.join((dbn1, mrna1.id, mrna1.stop - mrna1.start + 1))

        print(gfeats[gene], file=fw)

        for group in g:
            group.sort(key=lambda x: x[2], reverse=True)
            dbs, mrnas = [el[0] for el in group], [el[1] for el in group]
            d, m = dbs[0], mrnas[0]

            dbid, _mrnaid = "|".join(str(x) for x in set(dbs)), []
            for x in mrnas:
                if x not in _mrnaid:
                    _mrnaid.append(x)
            mrnaid = "{0}|{1}".format(dbid, "-".join(_mrnaid))
            if mrnaid not in seen:
                seen[mrnaid] = 0
            else:
                seen[mrnaid] += 1
                mrnaid = "{0}-{1}".format(mrnaid, seen[mrnaid])

            _mrna = gffdbx[d][m]
            _mrna.attributes["ID"] = [mrnaid]
            _mrna.attributes["Parent"] = [gene]
            children = gffdbx[d].children(m, order_by="start")
            print(_mrna, file=fw)
            for child in children:
                child.attributes["ID"] = ["{0}|{1}".format(dbid, child.id)]
                child.attributes["Parent"] = [mrnaid]
                print(child, file=fw)

            if opts.summary:
                summary = [mrnaid]
                summary.extend(["Y" if db in set(dbs) else "N" for db in gffdbx])
                print("\t".join(str(x) for x in summary), file=sfw)

            if opts.clusters:
                clusters = [mrnaid]
                clusters.append(",".join(str(el[0]) for el in group))
                clusters.append(",".join(str(el[1]) for el in group))
                clusters.append(",".join(str(el[2]) for el in group))
                print("\t".join(str(x) for x in clusters), file=cfw)

    fw.close()
    if opts.summary:
        sfw.close()
    if opts.clusters:
        cfw.close()


if __name__ == "__main__":
    main()



# cli.py
import argparse

from . import __version__


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    args = parser.parse_args()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
Array of data parsers for bioinformatics file formats, such as: GFF3, BED, SAM/BAM, VCF, PSL, AGP, FASTA/FASTQ, BLAST, etc.
"""

from ..apps.base import dmain

if __name__ == "__main__":
    dmain(__file__)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

from collections import defaultdict
from itertools import combinations, product
import os.path as op
import string
import sys

from ..apps.align import blast_main, diamond_blastp_main
from ..apps.align import last as last_main
from ..apps.base import (
    ActionDispatcher,
    OptionParser,
    glob,
    logger,
    mkdir,
    need_update,
    sh,
)
from ..compara.blastfilter import main as blastfilter_main
from ..compara.quota import main as quota_main
from ..compara.synteny import liftover, mcscan, scan
from ..formats.base import BaseFile, DictFile, must_open
from ..formats.bed import Bed
from ..formats.blast import (
    BlastLine,
    cscore,
)
from ..formats.blast import (
    filtered_blastfile_name,
)
from ..formats.blast import filter as blast_filter
from ..formats.fasta import Fasta
from ..utils.cbook import gene_name
from ..utils.grouper import Grouper
from .base import AnchorFile
from .synteny import check_beds


class OMGFile(BaseFile):
    def __init__(self, filename):
        super().__init__(filename)
        fp = open(filename)
        inblock = False
        components = []
        component = []
        for row in fp:
            if inblock:
                atoms = row.split()
                natoms = len(atoms)
                assert natoms in (0, 7)
                if natoms:
                    gene, taxa = atoms[0], atoms[5]
                    component.append((gene, taxa))
                else:
                    inblock = False
                    components.append(tuple(component))

            if row.strip().startswith("---"):
                inblock = True
                component = []

        if inblock:
            components.append(tuple(component))
        self.components = components

    def best(self):
        bb = set()
        for component in self.components:
            size = len(component)
            if size > 1:
                bb.add(component)
        return bb


def main():
    actions = (
        ("tandem", "identify tandem gene groups within certain distance"),
        ("ortholog", "run a combined synteny and RBH pipeline to call orthologs"),
        ("group", "cluster the anchors into ortho-groups"),
        ("omgprepare", "prepare weights file to run Sankoff OMG algorithm"),
        ("omg", "generate a series of Sankoff OMG algorithm inputs"),
        ("omgparse", "parse the OMG outputs to get gene lists"),
        ("enrich", "enrich OMG output by pulling genes missed by OMG"),
        ("layout", "layout the gene lists"),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def get_weights(weightsfiles=None):
    if weightsfiles is None:
        weightsfiles = glob("*.weights")

    weights = defaultdict(list)
    for row in must_open(weightsfiles):
        a, b, c = row.split()
        weights[a].append((a, b, c))
    return weights


def get_edges(weightsfiles=None):
    if weightsfiles is None:
        weightsfiles = glob("*.weights")

    edges = {}
    for row in must_open(weightsfiles):
        a, b, c = row.split()
        c = int(c)
        edges[(a, b)] = c
        edges[(b, a)] = c
    return edges


def get_info():
    infofiles = glob("*.info")
    info = {}
    for row in must_open(infofiles):
        a = row.split()[0]
        info[a] = row.rstrip()
    return info


def enrich(args):
    """
    %prog enrich omgfile groups ntaxa > enriched.omg

    Enrich OMG output by pulling genes misses by OMG.
    """
    p = OptionParser(enrich.__doc__)
    p.add_argument(
        "--ghost",
        default=False,
        action="store_true",
        help="Add ghost homologs already used",
    )
    opts, args = p.parse_args(args)

    if len(args) != 3:
        sys.exit(not p.print_help())

    omgfile, groupsfile, ntaxa = args
    ntaxa = int(ntaxa)
    ghost = opts.ghost

    # Get gene pair => weight mapping
    weights = get_edges()
    info = get_info()
    # Get gene => taxon mapping
    info = dict((k, v.split()[5]) for k, v in info.items())

    groups = Grouper()

    fp = open(groupsfile)
    for row in fp:
        members = row.strip().split(",")
        groups.join(*members)

    logger.debug(
        "Imported %d families with %d members.", len(groups), groups.num_members
    )

    seen = set()
    omggroups = Grouper()
    fp = open(omgfile)
    for row in fp:
        genes, idxs = row.split()
        genes = genes.split(",")
        seen.update(genes)
        omggroups.join(*genes)

    nmembers = omggroups.num_members
    logger.debug("Imported %d OMG families with %d members.", len(omggroups), nmembers)
    assert nmembers == len(seen)

    alltaxa = set(str(x) for x in range(ntaxa))
    recruited = []
    fp = open(omgfile)
    for row in fp:
        genes, idxs = row.split()
        genes = genes.split(",")
        a = genes[0]

        idxs = set(idxs.split(","))
        missing_taxa = alltaxa - idxs
        if not missing_taxa:
            print(row.rstrip())
            continue

        leftover = groups[a]
        if not ghost:
            leftover = set(leftover) - seen

        if not leftover:
            print(row.rstrip())
            continue

        leftover_sorted_by_taxa = dict(
            (k, [x for x in leftover if info[x] == k]) for k in missing_taxa
        )

        solutions = []
        for solution in product(*leftover_sorted_by_taxa.values()):
            score = sum(weights.get((a, b), 0) for a in solution for b in genes)
            if score == 0:
                continue
            score += sum(weights.get((a, b), 0) for a, b in combinations(solution, 2))
            solutions.append((score, solution))
            # print solution, score

        best_solution = max(solutions) if solutions else None
        if best_solution is None:
            print(row.rstrip())
            continue

        # print "best ==>", best_solution
        best_score, best_addition = best_solution
        genes.extend(best_addition)
        recruited.extend(best_addition)

        genes = sorted([(info[x], x) for x in genes])
        idxs, genes = zip(*genes)

        if ghost:  # decorate additions so it's clear that they were added
            pgenes = []
            for g in genes:
                if g in recruited and g in seen:
                    pgenes.append("|{0}|".format(g))
                else:
                    pgenes.append(g)
            genes = pgenes

        print("\t".join((",".join(genes), ",".join(idxs))))
        if not ghost:
            seen.update(best_addition)

    logger.debug("Recruited %d new genes.", len(recruited))


def pairwise_distance(a, b, threadorder):
    d = 0
    for x, y in zip(a, b)[:-1]:  # Last column not used
        x, y = x.strip("|"), y.strip("|")
        if "." in (x, y):
            dd = 50
        else:
            xi, x = threadorder[x]
            yi, y = threadorder[y]
            dd = min(abs(xi - yi), 50)
        d += dd
    return d


def insert_into_threaded(atoms, threaded, threadorder):
    min_idx, min_d = 0, 1000
    for i, t in enumerate(threaded):
        # calculate distance
        d = pairwise_distance(atoms, t, threadorder)
        if d < min_d:
            min_idx = i
            min_d = d

    i = min_idx
    t = threaded[i]
    threaded.insert(i, atoms)
    logger.debug("Insert %s before %s (d=%d)", atoms, t, min_d)


def sort_layout(thread, listfile, column=0):
    """
    Sort the syntelog table according to chromomomal positions. First orient the
    contents against threadbed, then for contents not in threadbed, insert to
    the nearest neighbor.
    """
    outfile = listfile.rsplit(".", 1)[0] + ".sorted.list"
    threadorder = thread.order
    fw = open(outfile, "w")
    lt = DictFile(listfile, keypos=column, valuepos=None)
    threaded = []
    imported = set()
    for t in thread:
        accn = t.accn
        if accn not in lt:
            continue

        imported.add(accn)
        atoms = lt[accn]
        threaded.append(atoms)

    assert len(threaded) == len(imported)

    total = sum(1 for x in open(listfile))
    logger.debug("Total: %d, currently threaded: %d", total, len(threaded))
    fp = open(listfile)
    for row in fp:
        atoms = row.split()
        accn = atoms[0]
        if accn in imported:
            continue
        insert_into_threaded(atoms, threaded, threadorder)

    for atoms in threaded:
        print("\t".join(atoms), file=fw)

    fw.close()
    logger.debug("File `%s` sorted to `%s`.", outfile, thread.filename)


def layout(args):
    """
    %prog layout omgfile taxa

    Build column formatted gene lists after omgparse(). Use species list
    separated by comma in place of taxa, e.g. "BR,BO,AN,CN"
    """
    p = OptionParser(layout.__doc__)
    p.add_argument("--sort", help="Sort layout file based on bedfile")
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    omgfile, taxa = args
    listfile = omgfile.rsplit(".", 1)[0] + ".list"
    taxa = taxa.split(",")
    ntaxa = len(taxa)
    fw = open(listfile, "w")

    data = []
    fp = open(omgfile)
    for row in fp:
        genes, idxs = row.split()
        row = ["."] * ntaxa
        genes = genes.split(",")
        ixs = [int(x) for x in idxs.split(",")]
        for gene, idx in zip(genes, ixs):
            row[idx] = gene
        txs = ",".join(taxa[x] for x in ixs)
        print("\t".join(("\t".join(row), txs)), file=fw)
        data.append(row)

    coldata = zip(*data)
    ngenes = []
    for i, tx in enumerate(taxa):
        genes = [x for x in coldata[i] if x != "."]
        genes = set(x.strip("|") for x in genes)
        ngenes.append((len(genes), tx))

    details = ", ".join("{0} {1}".format(a, b) for a, b in ngenes)
    total = sum(a for a, b in ngenes)
    s = "A list of {0} orthologous families that collectively".format(len(data))
    s += " contain a total of {0} genes ({1})".format(total, details)
    print(s, file=sys.stderr)

    fw.close()
    lastcolumn = ntaxa + 1
    cmd = "sort -k{0},{0} {1} -o {1}".format(lastcolumn, listfile)
    sh(cmd)

    logger.debug("List file written to `%s`.", listfile)
    sort = opts.sort
    if sort:
        thread = Bed(sort)
        sort_layout(thread, listfile)


def omgparse(args):
    """
    %prog omgparse work

    Parse the OMG outputs to get gene lists.
    """
    p = OptionParser(omgparse.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (work,) = args
    omgfiles = glob(op.join(work, "gf*.out"))
    for omgfile in omgfiles:
        omg = OMGFile(omgfile)
        best = omg.best()
        for bb in best:
            genes, taxa = zip(*bb)
            print("\t".join((",".join(genes), ",".join(taxa))))


def group(args):
    """
    %prog group anchorfiles

    Group the anchors into ortho-groups. Can input multiple anchor files.
    """
    p = OptionParser(group.__doc__)
    p.set_outfile()

    opts, args = p.parse_args(args)

    if len(args) < 1:
        sys.exit(not p.print_help())

    anchorfiles = args
    groups = Grouper()
    for anchorfile in anchorfiles:
        ac = AnchorFile(anchorfile)
        for a, b, idx in ac.iter_pairs():
            groups.join(a, b)

    logger.debug("Created %d groups with %d members.", len(groups), groups.num_members)

    outfile = opts.outfile
    fw = must_open(outfile, "w")
    for g in groups:
        print(",".join(sorted(g)), file=fw)
    fw.close()

    return outfile


def omg(args):
    """
    %prog omg weightsfile

    Run Sankoff's OMG algorithm to get orthologs. Download OMG code at:
    <http://137.122.149.195/IsbraSoftware/OMGMec.html>

    This script only writes the partitions, but not launch OMGMec. You may need to:

    $ parallel "java -cp ~/code/OMGMec TestOMGMec {} 4 > {}.out" ::: work/gf?????

    Then followed by omgparse() to get the gene lists.
    """
    p = OptionParser(omg.__doc__)

    opts, args = p.parse_args(args)

    if len(args) < 1:
        sys.exit(not p.print_help())

    weightsfiles = args
    groupfile = group(weightsfiles + ["--outfile=groups"])

    weights = get_weights(weightsfiles)
    info = get_info()

    fp = open(groupfile)

    work = "work"
    mkdir(work)
    for i, row in enumerate(fp):
        gf = op.join(work, "gf{0:05d}".format(i))
        genes = row.rstrip().split(",")

        fw = open(gf, "w")
        contents = ""
        npairs = 0
        for gene in genes:
            gene_pairs = weights[gene]
            for a, b, c in gene_pairs:
                if b not in genes:
                    continue

                contents += "weight {0}".format(c) + "\n"
                contents += info[a] + "\n"
                contents += info[b] + "\n\n"
                npairs += 1

        header = "a group of genes  :length ={0}".format(npairs)
        print(header, file=fw)
        print(contents, file=fw)

        fw.close()


def geneinfo(bed, genomeidx, ploidy):
    bedfile = bed.filename
    p = bedfile.split(".")[0]
    idx = genomeidx[p]
    pd = ploidy[p]
    infofile = p + ".info"

    if not need_update(bedfile, infofile):
        return infofile

    fwinfo = open(infofile, "w")

    for s in bed:
        chr = "".join(x for x in s.seqid if x in string.digits)
        try:
            chr = int(chr)
        except ValueError:
            chr = "0"

        print(
            "\t".join(str(x) for x in (s.accn, chr, s.start, s.end, s.strand, idx, pd)),
            file=fwinfo,
        )
    fwinfo.close()

    logger.debug("Update info file `%s`.", infofile)

    return infofile


def omgprepare(args):
    """
    %prog omgprepare ploidy anchorsfile blastfile

    Prepare to run Sankoff's OMG algorithm to get orthologs.
    """
    p = OptionParser(omgprepare.__doc__)
    p.add_argument("--norbh", action="store_true", help="Disable RBH hits")
    p.add_argument(
        "--pctid", default=0, type=int, help="Percent id cutoff for RBH hits"
    )
    p.add_argument("--cscore", default=90, type=int, help="C-score cutoff for RBH hits")
    p.set_stripnames()
    p.set_beds()

    opts, args = p.parse_args(args)

    if len(args) != 3:
        sys.exit(not p.print_help())

    ploidy, anchorfile, blastfile = args
    norbh = opts.norbh
    pctid = opts.pctid
    cs = opts.cscore
    qbed, sbed, qorder, sorder, is_self = check_beds(anchorfile, p, opts)

    fp = open(ploidy)
    genomeidx = dict((x.split()[0], i) for i, x in enumerate(fp))
    fp.close()

    ploidy = DictFile(ploidy)

    geneinfo(qbed, genomeidx, ploidy)
    geneinfo(sbed, genomeidx, ploidy)

    pf = blastfile.rsplit(".", 1)[0]
    cscorefile = pf + ".cscore"
    cscore([blastfile, "-o", cscorefile, "--cutoff=0", "--pct"])
    ac = AnchorFile(anchorfile)
    pairs = set((a, b) for a, b, i in ac.iter_pairs())
    logger.debug("Imported %d pairs from `%s`.", len(pairs), anchorfile)

    weightsfile = pf + ".weights"
    fp = open(cscorefile)
    fw = open(weightsfile, "w")
    npairs = 0
    for row in fp:
        a, b, c, pct = row.split()
        c, pct = float(c), float(pct)
        c = int(c * 100)
        if (a, b) not in pairs:
            if norbh:
                continue
            if c < cs:
                continue
            if pct < pctid:
                continue
            c /= 10  # This severely penalizes RBH against synteny

        print("\t".join((a, b, str(c))), file=fw)
        npairs += 1
    fw.close()

    logger.debug("Write %d pairs to `%s`.", npairs, weightsfile)


def make_ortholog(blocksfile, rbhfile, orthofile):
    # Generate mapping both ways
    adict = DictFile(rbhfile)
    bdict = DictFile(rbhfile, keypos=1, valuepos=0)
    adict.update(bdict)

    fp = open(blocksfile)
    fw = open(orthofile, "w")
    nrecruited = 0
    for row in fp:
        a, b = row.split()
        if b == ".":
            if a in adict:
                b = adict[a]
                nrecruited += 1
                b += "'"
        print("\t".join((a, b)), file=fw)

    logger.debug("Recruited %d pairs from RBH.", nrecruited)
    fp.close()
    fw.close()


def ortholog(args):
    """
    %prog ortholog species_a species_b

    Run a sensitive pipeline to find orthologs between two species a and b.
    The pipeline runs LAST and generate .lifted.anchors.

    `--full` mode would assume 1-to-1 quota synteny blocks as the backbone of
    such predictions. Extra orthologs will be recruited from reciprocal best
    match (RBH).
    """
    p = OptionParser(ortholog.__doc__)
    p.add_argument(
        "--dbtype",
        default="nucl",
        choices=("nucl", "prot"),
        help="Molecule type of subject database",
    )

    p.add_argument(
        "--full",
        default=False,
        action="store_true",
        help="Run in full 1x1 mode, including blocks and RBH",
    )
    p.add_argument(
        "--tandem_Nmax",
        type=int,
        default=10,
        help="merge tandem genes within distance",
    )
    p.add_argument("--cscore", default=0.7, type=float, help="C-score cutoff")
    p.add_argument(
        "--dist", default=20, type=int, help="Extent of flanking regions to search"
    )
    p.add_argument(
        "-n",
        "--min_size",
        dest="n",
        type=int,
        default=4,
        help="minimum number of anchors in a cluster",
    )
    p.add_argument("--quota", help="Quota align parameter")
    p.add_argument("--exclude", help="Remove anchors from a previous run")
    p.add_argument(
        "--self_remove",
        default=98,
        type=float,
        help="Remove self hits that are above this percent identity",
    )
    p.add_argument(
        "--no_strip_names",
        default=False,
        action="store_true",
        help="Do not strip alternative splicing (e.g. At5g06540.1 -> At5g06540)",
    )
    p.add_argument(
        "--liftover_dist",
        type=int,
        help="Distance to extend from liftover. Defaults to half of --dist",
    )
    p.set_cpus()
    dotplot_group = p.set_dotplot_opts()
    dotplot_group.add_argument(
        "--notex", default=False, action="store_true", help="Do not use tex"
    )
    dotplot_group.add_argument(
        "--no_dotplot", default=False, action="store_true", help="Do not make dotplot"
    )
    p.add_argument(
        "--ignore_zero_anchor",
        default=False,
        action="store_true",
        help="Ignore this pair of ortholog identification instead of throwing an error when performing many pairs of cataloging.",
    )
    p.add_argument(
        "--align_soft",
        default="last",
        choices=("last", "blast", "diamond_blastp"),
        help="Sequence alignment software. Default <last> for both <nucl> and <prot>. Users could also use <blast> for both <nucl> and <prot>, or <diamond_blastp> for <prot>.",
    )

    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    a, b = args
    dbtype = opts.dbtype
    ignore_zero_anchor = opts.ignore_zero_anchor
    suffix = ".cds" if dbtype == "nucl" else ".pep"
    abed, afasta = a + ".bed", a + suffix
    bbed, bfasta = b + ".bed", b + suffix
    ccscore = opts.cscore
    quota = opts.quota
    exclude = opts.exclude
    dist = "--dist={0}".format(opts.dist)
    minsize_flag = "--min_size={}".format(opts.n)
    cpus_flag = "--cpus={}".format(opts.cpus)
    align_soft = opts.align_soft

    aprefix = op.basename(a)
    bprefix = op.basename(b)
    pprefix = ".".join((aprefix, bprefix))
    qprefix = ".".join((bprefix, aprefix))
    last = pprefix + ".last"
    if need_update((afasta, bfasta), last, warn=True):
        if align_soft == "blast":
            blast_main([bfasta, afasta, cpus_flag], dbtype)
        elif dbtype == "prot" and align_soft == "diamond_blastp":
            diamond_blastp_main([bfasta, afasta, cpus_flag], dbtype)
        else:
            last_main([bfasta, afasta, cpus_flag], dbtype)

    self_remove = opts.self_remove
    if a == b:
        lastself = filtered_blastfile_name(last, self_remove, 0, inverse=True)
        if need_update(last, lastself, warn=True):
            blast_filter(
                [last, "--hitlen=0", f"--pctid={self_remove}", "--inverse", "--noself"]
            )
        last = lastself

    filtered_last = last + ".filtered"
    if need_update(last, filtered_last, warn=True):
        dargs = [
            last,
            f"--cscore={ccscore}",
            f"--tandem_Nmax={opts.tandem_Nmax}",
        ]
        if exclude:
            dargs += ["--exclude={}".format(exclude)]
        if opts.no_strip_names:
            dargs += ["--no_strip_names"]
        blastfilter_main(dargs)

    anchors = pprefix + ".anchors"
    lifted_anchors = pprefix + ".lifted.anchors"
    pdf = pprefix + ".pdf"
    if not opts.full:
        if need_update(filtered_last, lifted_anchors, warn=True):
            dargs = [
                filtered_last,
                anchors,
                minsize_flag,
                dist,
                "--liftover={0}".format(last),
            ]
            if opts.no_strip_names:
                dargs += ["--no_strip_names"]
            if opts.liftover_dist:
                dargs += ["--liftover_dist={}".format(opts.liftover_dist)]
            try:
                scan(dargs)
            except ValueError as e:
                if ignore_zero_anchor:
                    logger.debug(str(e))
                    logger.debug("Ignoring this error and continuing...")
                    return
                else:
                    raise ValueError(e) from e
        if quota:
            quota_main([lifted_anchors, "--quota={0}".format(quota), "--screen"])
        if need_update(anchors, pdf, warn=True) and not opts.no_dotplot:
            from jcvi.graphics.dotplot import dotplot_main

            dargs = [anchors]
            if opts.nostdpf:
                dargs += ["--nostdpf"]
            if opts.nochpf:
                dargs += ["--nochpf"]
            if opts.skipempty:
                dargs += ["--skipempty"]
            if opts.genomenames:
                dargs += ["--genomenames", opts.genomenames]
            if opts.theme:
                dargs += ["--theme", opts.theme]
            if opts.notex:
                dargs += ["--notex"]
            dotplot_main(dargs)
        return

    if need_update(filtered_last, anchors, warn=True):
        if opts.no_strip_names:
            scan([filtered_last, anchors, dist, "--no_strip_names"])
        else:
            scan([filtered_last, anchors, dist])

    ooanchors = pprefix + ".1x1.anchors"
    if need_update(anchors, ooanchors, warn=True):
        quota_main([anchors, "--quota=1:1", "--screen"])

    lifted_anchors = pprefix + ".1x1.lifted.anchors"
    if need_update((last, ooanchors), lifted_anchors, warn=True):
        if opts.no_strip_names:
            liftover([last, ooanchors, dist, "--no_strip_names"])
        else:
            liftover([last, ooanchors, dist])

    pblocks = pprefix + ".1x1.blocks"
    qblocks = qprefix + ".1x1.blocks"
    if need_update(lifted_anchors, [pblocks, qblocks], warn=True):
        mcscan([abed, lifted_anchors, "--iter=1", "-o", pblocks])
        mcscan([bbed, lifted_anchors, "--iter=1", "-o", qblocks])

    rbh = pprefix + ".rbh"
    if need_update(last, rbh, warn=True):
        cscore([last, "-o", rbh])

    portho = pprefix + ".ortholog"
    qortho = qprefix + ".ortholog"
    if need_update([pblocks, qblocks, rbh], [portho, qortho], warn=True):
        make_ortholog(pblocks, rbh, portho)
        make_ortholog(qblocks, rbh, qortho)


def tandem_main(
    blast_file,
    cds_file,
    bed_file,
    N=3,
    P=50,
    is_self=True,
    evalue=0.01,
    strip_name=".",
    ofile=sys.stderr,
    genefam=False,
):
    if genefam:
        N = 1e5

    # get the sizes for the CDS first
    f = Fasta(cds_file)
    sizes = dict(f.itersizes())

    # retrieve the locations
    bed = Bed(bed_file)
    order = bed.order

    if is_self:
        # filter the blast file
        g = Grouper()
        fp = open(blast_file)
        for row in fp:
            b = BlastLine(row)
            query_len = sizes[b.query]
            subject_len = sizes[b.subject]
            if b.hitlen < min(query_len, subject_len) * P / 100.0:
                continue

            query = gene_name(b.query, sep=strip_name)
            subject = gene_name(b.subject, sep=strip_name)
            qi, q = order[query]
            si, s = order[subject]

            if abs(qi - si) <= N and b.evalue <= evalue:
                if genefam:
                    g.join(query, subject)
                elif q.seqid == s.seqid:
                    g.join(query, subject)

    else:
        homologs = Grouper()
        fp = open(blast_file)
        for row in fp:
            b = BlastLine(row)
            query_len = sizes[b.query]
            subject_len = sizes[b.subject]
            if b.hitlen < min(query_len, subject_len) * P / 100.0:
                continue
            if b.evalue > evalue:
                continue

            query = gene_name(b.query, sep=strip_name)
            subject = gene_name(b.subject, sep=strip_name)
            homologs.join(query, subject)

        if genefam:
            g = homologs
        else:
            g = Grouper()
            for i, atom in enumerate(bed):
                for x in range(1, N + 1):
                    if all(
                        [
                            i - x >= 0,
                            bed[i - x].seqid == atom.seqid,
                            homologs.joined(bed[i - x].accn, atom.accn),
                        ]
                    ):
                        leni = sizes[bed[i].accn]
                        lenx = sizes[bed[i - x].accn]
                        if abs(leni - lenx) > max(leni, lenx) * (1 - P / 100.0):
                            continue
                        g.join(bed[i - x].accn, atom.accn)

    # dump the grouper
    fw = must_open(ofile, "w")
    ngenes, nfamilies = 0, 0
    families = []
    for group in sorted(g):
        if len(group) >= 2:
            print(",".join(sorted(group)), file=fw)
            ngenes += len(group)
            nfamilies += 1
            families.append(sorted(group))

    longest_family = max(families, key=lambda x: len(x))

    # generate reports
    print("Proximal paralogues (dist=%d):" % N, file=sys.stderr)
    print("Total %d genes in %d families" % (ngenes, nfamilies), file=sys.stderr)
    print(
        "Longest families (%d): %s" % (len(longest_family), ",".join(longest_family)),
        file=sys.stderr,
    )

    return families


def tandem(args):
    """
    %prog tandem blast_file cds_file bed_file [options]

    Find tandem gene clusters that are separated by N genes, based on filtered
    blast_file by enforcing alignments between any two genes at least 50%
    (or user specified value) of either gene.

    pep_file can also be used in same manner.
    """
    p = OptionParser(tandem.__doc__)
    p.add_argument(
        "--tandem_Nmax",
        dest="tandem_Nmax",
        type=int,
        default=3,
        help="merge tandem genes within distance",
    )
    p.add_argument(
        "--percent_overlap",
        type=int,
        default=50,
        help="tandem genes have >=x% aligned sequence, x=0-100",
    )
    p.set_align(evalue=0.01)
    p.add_argument(
        "--not_self",
        default=False,
        action="store_true",
        help="provided is not self blast file",
    )
    p.add_argument(
        "--strip_gene_name",
        dest="sep",
        type=str,
        default=".",
        help="strip alternative splicing. Use None for no stripping.",
    )
    p.add_argument(
        "--genefamily",
        dest="genefam",
        action="store_true",
        help="compile gene families based on similarity",
    )
    p.set_outfile()

    opts, args = p.parse_args(args)

    if len(args) != 3:
        sys.exit(not p.print_help())

    blast_file, cds_file, bed_file = args
    N = opts.tandem_Nmax
    P = opts.percent_overlap
    is_self = not opts.not_self
    sep = opts.sep
    ofile = opts.outfile

    tandem_main(
        blast_file,
        cds_file,
        bed_file,
        N=N,
        P=P,
        is_self=is_self,
        evalue=opts.evalue,
        strip_name=sep,
        ofile=ofile,
        genefam=opts.genefam,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Convert PAF/LASTZ alignments to bed and block files for jcvi.ribbon
using layout file as guide.
"""

import argparse
from collections import namedtuple
import glob
from operator import attrgetter
import os
import shutil
import sys

from jcvi.graphics.base import AbstractLayout


def mainArgs():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Convert whole-genome alignments to block & bed format \
                    for ribbon.",
        prog="jvci.apps.wga2ribbon",
    )
    parser.add_argument(
        "-f",
        "--format",
        type=str,
        required=True,
        choices=["paf", "lastz"],
        help="Set alignment format as paf or lastz.",
    )
    parser.add_argument(
        "-a",
        "--alignments",
        action="append",
        required=True,
        help='Set arg for each alignment file to be imported.\
                        Has format "[Source chrm Idx],[Hit chrm Idx],\
                        [filename]" i.e. For file with alignmnets between the \
                        first and second tracks in ribbon.layout: \
                        -a "0,1,A2B_Alignment.tab"',
    )
    parser.add_argument(
        "-l", "--layout", type=str, required=True, help="Ribbon plot layout file."
    )
    parser.add_argument(
        "--minID",
        type=float,
        default=50,
        help="Extract alignments with identity => than minID",
    )
    parser.add_argument(
        "--outBed",
        type=str,
        default="ribbons.bed",
        help="Write alignment block coords to this bedfile",
    )
    parser.add_argument(
        "--outBlocks",
        type=str,
        default="ribbons.block",
        help="Write alignment block names to this file.",
    )
    args = parser.parse_args()
    return args


class LayoutLine(object):
    #  0-11: x,y,rotation,ha,va,color,ratio,label,chrmName,rStart,rEnd,chrmMax
    def __init__(self, row, delimiter=","):
        self.hidden = row[0] == "*"
        if self.hidden:
            row = row[1:]
        args = row.rstrip().split(delimiter)
        args = [x.strip() for x in args]
        self.x = float(args[0])
        self.y = float(args[1])
        self.rotation = int(args[2])
        self.ha = args[3]
        self.va = args[4]
        self.color = args[5]
        self.ratio = 1
        if len(args) > 6:
            self.ratio = float(args[6])
        if len(args) > 7:
            self.label = args[7].strip()
        else:
            self.label = None
        if len(args) > 8:
            self.ChrmName = str(args[8])
        if len(args) > 9:
            self.rStart = int(args[9])
            self.rEnd = int(args[10])
        else:
            self.rStart = None
            self.rEnd = None
        if len(args) > 10:
            self.chrmMax = int(args[10])
        else:
            self.chrmMax = None


class Layout(AbstractLayout):
    def __init__(self, filename, delimiter=","):
        super(Layout, self).__init__(filename)
        fp = open(filename)
        self.edges = []
        for row in fp:
            # Skip blank lines
            if row.strip():
                # Skip header rows
                if row[0] == "#":
                    continue
                # Import edges
                if row[0] == "e":
                    args = row.rstrip().split(delimiter)
                    args = [x.strip() for x in args]
                    # From index 1 and 2
                    a, b = args[1:3]
                    a, b = int(a), int(b)
                    assert args[0] == "e"
                    self.edges.append((a, b))
                else:
                    # Parse all other lines as sequence tracks
                    self.append(LayoutLine(row, delimiter=delimiter))

        # Track number of layout lines added
        self.lines = len(self)

        if 3 <= len(self) <= 8:
            self.assign_colors()


def readLASTZ(infile, minID=50):
    """
    Read in LASTZ result file from LASTZ_genome_align.sh
    Populate nested dictionary of hits keyed by Target
    and then Query scaffold names.
    # Check diff between standard LASTZ and mimeo-LASTZ formats
    """
    with open(infile) as f:
        content = f.readlines()
    content = [x.strip().split() for x in content]
    hitsDict = dict()
    # Set named tuple format
    hitTup = namedtuple(
        "Elem",
        [
            "t_start",
            "t_end",
            "t_strand",
            "q_start",
            "q_end",
            "q_strand",
            "idPct",
            "UID",
        ],
    )
    counter = 0
    # Read in split rows
    for row in content:
        # Ignore lines begining with '#'
        if row[0][0] == "#":
            continue
        # Count if hit identity exceeds threshold
        # CS10_Chromosome_08    +    1993504    1993793    supercont1.32    +    15987    16251    8424    72.9
        elif float(row[9]) >= minID:
            counter += 1
            UID = counter
            t_name = str(row[0])
            t_strand = str(row[1])
            t_start = int(row[2])  # - 1 # Convert from idx '1' to idx '0'
            t_end = int(row[3])  # - 1 # Convert from idx '1' to idx '0'
            q_name = str(row[4])
            q_strand = str(row[5])
            idPct = float(row[9])
            # Check that start position < end position
            if int(row[6]) < int(row[7]):
                q_start = int(row[6])  # - 1 # Convert from idx '1' to idx '0'
                q_end = int(row[7])  # - 1 # Convert from idx '1' to idx '0'
            # Correct for inverted coordinates
            else:
                print(("Inverting query sequence coordinates for record number:", UID))
                q_end = int(row[6])  # - 1 # Convert from idx '1' to idx '0'
                q_start = int(row[7])  # - 1 # Convert from idx '1' to idx '0'
            # Create target scaffold dict if not seen
            if t_name not in list(hitsDict.keys()):
                hitsDict[t_name] = dict()
            # Create query scaffold list if not seen
            if q_name not in list(hitsDict[t_name].keys()):
                hitsDict[t_name][q_name] = list()
            # Write record to target:query list as named tuple
            hitsDict[t_name][q_name].append(
                hitTup(t_start, t_end, t_strand, q_start, q_end, q_strand, idPct, UID)
            )
    return hitsDict


def readPAF(infile, minID=50):
    """
    Format info here: https://github.com/lh3/miniasm/blob/master/PAF.md
    From PAF file populate nested dictionary of hits keyed by Target (Ref) and
    then Query scaffold names.
    # Check target and query seq defs for PAF and LASTZ
    # Check what PAF does with rev orientation coords - flipped?
    # Check if ribbon wants intervals indexed from zero
    # PAF is zero based
    # Check how strand orientation is assigned
    """
    # |Col|Type  |Description                               |
    # |--:|:----:|:-----------------------------------------|
    # |0  |string|Query sequence name                       |
    # |1  |int   |Query sequence length                     |
    # |2  |int   |Query start (0-based; BED-like; closed)   |
    # |3  |int   |Query end (0-based; BED-like; open)       |
    # |4  |char  |Relative strand: "+" or "-"               |
    # |5  |string|Target sequence name                      |
    # |6  |int   |Target sequence length                    |
    # |7  |int   |Target start on original strand (0-based) |
    # |8  |int   |Target end on original strand (0-based)   |
    # |9  |int   |Number of residue matches                 |
    # |10 |int   |Alignment block length                    |
    # |11 |int   |Mapping quality (0-255; 255 for missing)  |
    with open(infile) as f:
        content = f.readlines()
    content = [x.strip().split() for x in content]
    hitsDict = dict()
    # Set named tuple format
    hitTup = namedtuple(
        "Elem",
        [
            "t_start",
            "t_end",
            "t_strand",
            "q_start",
            "q_end",
            "q_strand",
            "idPct",
            "UID",
        ],
    )
    counter = 0
    # Read in split rows
    for row in content:
        # Ignore lines begining with '#'
        if row[0][0] == "#":
            continue
        # Count if hit identity exceeds threshold
        # For PAF calc idenity from matching bases and alignment length.
        rowID = float(row[9]) / float(row[10])
        if rowID >= minID:
            counter += 1
            UID = counter
            t_name = str(row[5])
            t_strand = "+"
            t_start = int(row[7])  # + 1 # Convert from idx '0' to idx '1'
            t_end = int(row[8])  # + 1 # Convert from idx '0' to idx '1'
            q_name = str(row[0])
            q_strand = str(row[4])
            idPct = rowID
            # Check that start position < end position
            if int(row[2]) < int(row[3]):
                q_start = int(row[2])  # + 1 # Convert from idx '0' to idx '1'
                q_end = int(row[3])  # + 1 # Convert from idx '0' to idx '1'
            # Correct for inverted coordinates
            else:
                print(("Inverting query sequence coordinates for record number: ", UID))
                q_end = int(row[2])  # + 1 # Convert from idx '0' to idx '1'
                q_start = int(row[3])  # + 1 # Convert from idx '0' to idx '1'
            # Create target scaffold dict if not seen
            if t_name not in list(hitsDict.keys()):
                hitsDict[t_name] = dict()
            # Create query scaffold list if not seen
            if q_name not in list(hitsDict[t_name].keys()):
                hitsDict[t_name][q_name] = list()
            # Write record to target:query list as named tuple
            hitsDict[t_name][q_name].append(
                hitTup(t_start, t_end, t_strand, q_start, q_end, q_strand, idPct, UID)
            )
    return hitsDict


def main():
    # Get cmd line args
    args = mainArgs()
    plotlines = Layout(args.layout)

    # Check number of alignment files = layout lines -1 (i.e. AvB.tab, BvC.tab)
    # if plotlines.lines != len(args.alignments) + 1:
    #    print('Error: Alignment file list must be == number sequence lines in layout - 1')
    #    sys.exit(0)

    # Init dict with layout row indexs as keys and empty list as value
    bedlines = dict()
    for i in range(plotlines.lines):
        bedlines[i] = list()
    # Init blocklines
    blocklines = list()
    for coords in args.alignments:
        assert len(coords.split(",")) == 3
        selfIdx, hitIdx, alnFile = coords.split(",")
        selfIdx = int(selfIdx)
        hitIdx = int(hitIdx)
        hits = readLASTZ(alnFile, minID=args.minID)
        # Extract hits between sequence selfIdx and hitIdx as per layout file.
        for x in sorted(
            hits[plotlines[selfIdx].ChrmName][plotlines[hitIdx].ChrmName],
            key=lambda x: (int(x.t_start), int(x.t_end)),
        ):
            tName = str(plotlines[selfIdx].ChrmName)
            qName = str(plotlines[hitIdx].ChrmName)
            tUID = "_".join(
                [
                    str(selfIdx),
                    str(hitIdx),
                    str(selfIdx),
                    "ID" + str(x.UID),
                    tName,
                    str(x.t_start),
                    str(x.t_end),
                ]
            )
            qUID = "_".join(
                [
                    str(selfIdx),
                    str(hitIdx),
                    str(hitIdx),
                    "ID" + str(x.UID),
                    qName,
                    str(x.q_start),
                    str(x.q_end),
                ]
            )
            bedlines[selfIdx].append(
                tuple(
                    [
                        tName,
                        str(x.t_start),
                        str(x.t_end),
                        tUID,
                        str(x.idPct),
                        str(x.t_strand),
                    ]
                )
            )
            bedlines[hitIdx].append(
                tuple(
                    [
                        qName,
                        str(x.q_start),
                        str(x.q_end),
                        qUID,
                        str(x.idPct),
                        str(x.q_strand),
                    ]
                )
            )
            blockline = ["."] * plotlines.lines
            blockline[selfIdx] = tUID
            blockline[hitIdx] = qUID
            blocklines.append("\t".join(blockline))

    # Sort bedlines for each seq in layout file.
    # Sort by Chrm name, start, end, UID
    with open(args.outBed, "w") as f:
        for layoutIdx in list(bedlines.keys()):
            for line in sorted(
                bedlines[layoutIdx], key=lambda x: (x[0], int(x[1]), int(x[2]), x[3])
            ):
                f.write("\t".join(line) + "\n")

    # Write blocklines
    with open(args.outBlocks, "w") as f:
        for line in blocklines:
            f.write(line + "\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""Package setup for Cython extensions only"""

from Cython.Build import build_ext
from setuptools import setup, Extension
import numpy as np

ext_modules = [
    Extension(
        "jcvi.assembly.chic",
        ["src/jcvi/assembly/chic.pyx"],
        include_dirs=[np.get_include()],
        extra_compile_args=["-O3"],
    ),
    Extension(
        "jcvi.formats.cblast",
        ["src/jcvi/formats/cblast.pyx"],
        extra_compile_args=["-O3"],
    ),
]

setup(
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
%prog blast_file --qbed query.bed --sbed subject.bed

Accepts bed format and blast file, and run several BLAST filters below::

* Local dup filter:
if the input is query.bed and subject.bed, the script files query.localdups
and subject.localdups are created containing the parent|offspring dups, as
inferred by subjects hitting same query or queries hitting same subject.

* C-score filter:
see supplementary info for sea anemone genome paper, formula::

    cscore(A,B) = score(A,B) /
         max(best score for A, best score for B)

Finally a blast.filtered file is created.
"""
from collections import defaultdict
from itertools import groupby
import os.path as op
import sys

from ..apps.base import OptionParser, logger
from ..compara.synteny import check_beds
from ..formats.blast import Blast
from ..utils.cbook import gene_name
from ..utils.grouper import Grouper


def blastfilter_main(blast_file, p, opts):

    qbed, sbed, qorder, sorder, is_self = check_beds(blast_file, p, opts)

    tandem_Nmax = opts.tandem_Nmax
    cscore = opts.cscore
    exclude = opts.exclude

    fp = open(blast_file)
    total_lines = sum(1 for line in fp if line[0] != "#")
    logger.debug(
        "Load BLAST file `{}` (total {} lines)".format(blast_file, total_lines)
    )
    bl = Blast(blast_file)
    blasts = sorted(list(bl), key=lambda b: b.score, reverse=True)

    filtered_blasts = []
    seen = set()
    ostrip = opts.strip_names
    nwarnings = 0
    for b in blasts:
        query, subject = b.query, b.subject
        if is_self and query == subject:
            continue

        if ostrip:
            query, subject = gene_name(query), gene_name(subject)
        if query not in qorder:
            if nwarnings < 100:
                logger.warning("{} not in {}".format(query, qbed.filename))
            elif nwarnings == 100:
                logger.warning("too many warnings.. suppressed")
            nwarnings += 1
            continue
        if subject not in sorder:
            if nwarnings < 100:
                logger.warning("{} not in {}".format(subject, sbed.filename))
            elif nwarnings == 100:
                logger.warning("too many warnings.. suppressed")
            nwarnings += 1
            continue

        qi, q = qorder[query]
        si, s = sorder[subject]

        if is_self and qi > si:
            # move all hits to same side when doing self-self BLAST
            query, subject = subject, query
            qi, si = si, qi
            q, s = s, q

        key = query, subject
        if key in seen:
            continue
        seen.add(key)
        b.query, b.subject = [str(k) for k in key]

        b.qi, b.si = qi, si
        b.qseqid, b.sseqid = q.seqid, s.seqid

        filtered_blasts.append(b)

    if exclude:
        before_filter = len(filtered_blasts)
        logger.debug("running excluded pairs (--exclude `{}`) ..".format(exclude))
        filtered_blasts = list(filter_exclude(filtered_blasts, exclude=exclude))
        logger.debug(
            "after filter ({}->{}) ..".format(before_filter, len(filtered_blasts))
        )

    if cscore:
        before_filter = len(filtered_blasts)
        logger.debug("running the cscore filter (cscore>=%.2f) .." % cscore)
        filtered_blasts = list(filter_cscore(filtered_blasts, cscore=cscore))
        logger.debug(
            "after filter ({}->{}) ..".format(before_filter, len(filtered_blasts))
        )

    if tandem_Nmax:
        logger.debug(
            "running the local dups filter (tandem_Nmax={}) ..".format(tandem_Nmax)
        )

        qtandems = tandem_grouper(filtered_blasts, flip=True, tandem_Nmax=tandem_Nmax)
        standems = tandem_grouper(filtered_blasts, flip=False, tandem_Nmax=tandem_Nmax)

        qdups_fh = (
            open(op.splitext(opts.qbed)[0] + ".localdups", "w")
            if opts.tandems_only
            else None
        )

        if is_self:
            for s in standems:
                qtandems.join(*s)
            qdups_to_mother = write_localdups(qtandems, qbed, qdups_fh)
            sdups_to_mother = qdups_to_mother
        else:
            qdups_to_mother = write_localdups(qtandems, qbed, qdups_fh)
            sdups_fh = (
                open(op.splitext(opts.sbed)[0] + ".localdups", "w")
                if opts.tandems_only
                else None
            )
            sdups_to_mother = write_localdups(standems, sbed, sdups_fh)

        if opts.tandems_only:
            # write out new .bed after tandem removal
            write_new_bed(qbed, qdups_to_mother)
            if not is_self:
                write_new_bed(sbed, sdups_to_mother)

            # just want to use this script as a tandem finder.
            # sys.exit()

        before_filter = len(filtered_blasts)
        filtered_blasts = list(
            filter_tandem(filtered_blasts, qdups_to_mother, sdups_to_mother)
        )
        logger.debug(
            "after filter ({}->{}) ..".format(before_filter, len(filtered_blasts))
        )

    blastfilteredfile = blast_file + ".filtered"
    fw = open(blastfilteredfile, "w")
    write_new_blast(filtered_blasts, fh=fw)
    fw.close()


def write_localdups(tandems, bed, dups_fh=None):

    tandem_groups = []
    for group in tandems:
        rows = [bed[i] for i in group]
        # within the tandem groups, genes are sorted with decreasing size
        rows.sort(key=lambda a: (-abs(a.end - a.start), a.accn))
        tandem_groups.append([x.accn for x in rows])

    dups_to_mother = {}
    n = 1
    for accns in sorted(tandem_groups):
        if dups_fh:
            print("\t".join(accns), file=dups_fh)
            if n:
                n -= 1
                logger.debug("write local dups to file {}".format(dups_fh.name))

        for dup in accns[1:]:
            dups_to_mother[dup] = accns[0]

    return dups_to_mother


def write_new_bed(bed, children):
    # generate local dup removed annotation files
    out_name = "%s.nolocaldups%s" % op.splitext(bed.filename)
    logger.debug("write tandem-filtered bed file %s" % out_name)
    fh = open(out_name, "w")
    for i, row in enumerate(bed):
        if row["accn"] in children:
            continue
        print(row, file=fh)
    fh.close()


def write_new_blast(filtered_blasts, fh=sys.stdout):
    for b in filtered_blasts:
        print(b, file=fh)


def filter_exclude(blast_list, exclude=None):
    """Filter gene pairs from an excluded list

    Args:
        blast_list (List[BlastLine]): List of BlastLines
        exclude (str, optional): Path to the excluded anchors file. Defaults to None.
    """
    from .base import AnchorFile

    excluded_pairs = set()
    ac = AnchorFile(exclude)
    for a, b, block in ac.iter_pairs():
        excluded_pairs.add((a, b))
        excluded_pairs.add((b, a))
    for b in blast_list:
        if (b.query, b.subject) in excluded_pairs:
            continue
        yield b


def filter_cscore(blast_list, cscore=0.5):

    best_score = defaultdict(float)
    for b in blast_list:
        if b.score > best_score[b.query]:
            best_score[b.query] = b.score
        if b.score > best_score[b.subject]:
            best_score[b.subject] = b.score

    for b in blast_list:
        cur_cscore = b.score / max(best_score[b.query], best_score[b.subject])
        if cur_cscore > cscore:
            yield b


def filter_tandem(blast_list, qdups_to_mother, sdups_to_mother):

    mother_blast = []
    for b in blast_list:
        if b.query in qdups_to_mother:
            b.query = qdups_to_mother[b.query]
        if b.subject in sdups_to_mother:
            b.subject = sdups_to_mother[b.subject]
        mother_blast.append(b)

    mother_blast.sort(key=lambda b: b.score, reverse=True)
    seen = {}
    for b in mother_blast:
        if b.query == b.subject:
            continue
        key = b.query, b.subject
        if key in seen:
            continue
        seen[key] = None
        yield b


def tandem_grouper(blast_list, tandem_Nmax=10, flip=True):
    if not flip:
        simple_blast = [
            (b.query, (b.sseqid, b.si)) for b in blast_list if b.evalue < 1e-10
        ]
    else:
        simple_blast = [
            (b.subject, (b.qseqid, b.qi)) for b in blast_list if b.evalue < 1e-10
        ]

    simple_blast.sort()

    standems = Grouper()
    for name, hits in groupby(simple_blast, key=lambda x: x[0]):
        # these are already sorted.
        hits = [x[1] for x in hits]
        for ia, a in enumerate(hits[:-1]):
            b = hits[ia + 1]
            # on the same chr and rank difference no larger than tandem_Nmax
            if b[1] - a[1] <= tandem_Nmax and b[0] == a[0]:
                standems.join(a[1], b[1])

    return standems


def main(args):

    p = OptionParser(__doc__)
    p.set_beds()
    p.set_stripnames()
    p.add_argument(
        "--tandems_only",
        dest="tandems_only",
        action="store_true",
        default=False,
        help="only calculate tandems, write .localdup file and exit.",
    )
    p.add_argument(
        "--tandem_Nmax",
        type=int,
        default=10,
        help="merge tandem genes within distance",
    )
    p.add_argument(
        "--cscore",
        type=float,
        default=0.7,
        help="retain hits that have good bitscore. a value of 0.5 means "
        "keep all values that are 50%% or greater of the best hit. "
        "higher is more stringent",
    )
    p.add_argument("--exclude", help="Remove anchors from a previous run")

    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (blastfile,) = args
    blastfilter_main(blastfile, p, opts)


if __name__ == "__main__":
    main(sys.argv[1:])
